"""Test command-line interface"""
import sys
import tempfile
import pathlib
import shutil

from dclab import cli, new_dataset, rtdc_dataset

import h5py
import numpy as np
import pytest

from helper_methods import retrieve_data


def test_check_suffix_disabled_repack():
    path_in_o = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_in = path_in_o.with_suffix("")
    path_in_o.rename(path_in)
    assert path_in.suffix == ""
    with pytest.raises(ValueError, match="Unsupported file type"):
        cli.repack(path_in=path_in,
                   path_out=path_in.with_name("repacked.rtdc"))
    # but this should work:
    cli.repack(path_in=path_in,
               path_out=path_in.with_name("repacked2.rtdc"),
               check_suffix=False)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("method", ["compress", "condense", "repack"])
def test_compressed(method):
    """Make sure the output data are compressed"""
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    mcallable = getattr(cli, method)
    mcallable(path_in=path_in,
              path_out=path_out)

    ic = rtdc_dataset.check.IntegrityChecker(path_out)
    ccue = ic.check_compression()[0]
    assert ccue.data["uncompressed"] == 0


def test_method_available():
    # DCOR depotize needs this
    assert hasattr(cli, "get_job_info")
    assert hasattr(cli, "get_command_log")


def test_tdms2rtdc():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=False)

    with new_dataset(path_out) as ds2, new_dataset(path_in) as ds1:
        assert len(ds2)
        assert set(ds1.features) == set(ds2.features)
        # not all features are computed
        assert set(ds2._events.keys()) < set(ds1.features)
        for feat in ds1:
            assert np.all(ds1[feat] == ds2[feat])


def test_tdms2rtdc_bulk():
    pytest.importorskip("nptdms")
    path_data = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    path_wd = pathlib.Path(
        tempfile.mkdtemp(prefix="tdms2rtdc_bulk_")).resolve()
    path_in = path_wd / "input"
    path_in.mkdir()
    shutil.copytree(path_data.parent, path_in / "data_1")
    shutil.copytree(path_data.parent, path_in / "data_2")
    shutil.copytree(path_data.parent, path_in / "data_3")
    (path_in / "data_nested").mkdir()
    shutil.copytree(path_data.parent, path_in / "data_nested" / "data_4")
    # same directory (will be cleaned up with path_in)
    path_out = path_wd / "output"
    path_out.mkdir()

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=False)

    for pp in [path_out / "data_1" / "M1_data.rtdc",
               path_out / "data_2" / "M1_data.rtdc",
               path_out / "data_3" / "M1_data.rtdc",
               path_out / "data_nested" / "data_4" / "M1_data.rtdc"]:
        assert pp.exists()

        with new_dataset(pp) as ds2, new_dataset(path_data) as ds1:
            assert len(ds2)
            assert set(ds1.features) == set(ds2.features)
            # not all features are computed
            assert set(ds2._events.keys()) < set(ds1.features)
            for feat in ds1:
                assert np.all(ds1[feat] == ds2[feat])


def test_tdms2rtdc_features():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=True)

    with new_dataset(path_out) as ds2, new_dataset(path_in) as ds1:
        assert len(ds2)
        assert set(ds1.features) == set(ds2.features)
        # features were computed
        assert set(ds2.features_loaded) == set(ds1.features)


def test_tdms2rtdc_remove_nan_image():
    pytest.importorskip("nptdms")
    imageio = pytest.importorskip("imageio")
    path_in = retrieve_data("fmt-tdms_fl-image-bright_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out_a = path_in.with_name("out_a.rtdc")
    path_out_b = path_in.with_name("out_b.rtdc")

    # generate fake video
    with new_dataset(path_in) as ds:
        video_length = len(ds) - 1
    vname = path_in.with_name("M4_0.040000ul_s_imaq.avi")
    # remove contour data (not necessary for this test)
    path_in.with_name("M4_0.040000ul_s_contours.txt").unlink()

    imgs = imageio.mimread(vname)
    with imageio.get_writer(vname, macro_block_size=1) as writer:
        for ii in range(video_length):
            writer.append_data(imgs[ii % len(imgs)])

    # without removal
    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out_a,
                  compute_features=False,
                  skip_initial_empty_image=False)

    with new_dataset(path_out_a) as ds2, new_dataset(path_in) as ds1:
        assert np.all(ds2["image"][0] == 0)
        assert len(ds2["image"]) == len(ds1)
        assert len(ds2) == len(ds1)

    # with removal
    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out_b,
                  compute_features=False,
                  skip_initial_empty_image=True)

    with new_dataset(path_out_b) as ds2:
        assert not np.all(ds2["image"][0] == 0)
        assert len(ds2) == video_length
        assert ds2.config["experiment"]["event count"] == video_length


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_tdms2rtdc_update_roi_size():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_fl-image_2016.zip")
    # set wrong roi sizes
    camin = path_in.with_name("M1_camera.ini")
    with camin.open("r") as fd:
        lines = fd.readlines()
    lines = lines[:-2]
    lines.append("width = 23\n")
    lines.append("height = 24\n")
    with camin.open("w") as fd:
        fd.writelines(lines)

    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=False,
                  skip_initial_empty_image=True)

    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert ds0.config["imaging"]["roi size x"] == 23
        assert ds0.config["imaging"]["roi size y"] == 24
        assert dsj.config["imaging"]["roi size x"] == 256
        assert dsj.config["imaging"]["roi size y"] == 96
        wlog = "dclab-tdms2rtdc-warnings"
        assert "LimitingExportSizeWarning" in dsj.logs[wlog]


def test_tdms2rtdc_update_sample_per_events():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_2fl-no-image_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    # set wrong samples per event
    with path_in.with_name("M1_para.ini").open("a") as fd:
        fd.write("Samples Per Event = 1234")

    with new_dataset(path_in) as ds:
        assert ds.config["fluorescence"]["samples per event"] == 1234

    cli.tdms2rtdc(path_tdms=path_in,
                  path_rtdc=path_out,
                  compute_features=True)

    with new_dataset(path_out) as ds2:
        assert ds2.config["fluorescence"]["samples per event"] == 566


@pytest.mark.parametrize("dataset,exit_status_expected", [
    ["fmt-hdf5_fl_2017.zip", 3],
    ["fmt-hdf5_fl_2018.zip", 1],
    ["fmt-hdf5_polygon_gate_2021.zip", 0],
])
def test_verify_dataset_exit_codes(dataset, exit_status_expected, monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data(dataset)
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == exit_status_expected


def test_verify_dataset_exit_code_alert(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["experiment:unknown"] = ""
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 1  # unknown key leads to alert


def test_verify_dataset_exit_code_error(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["setup:channel width"] = "peter"
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 4  # Cannot convert string to float error


def test_verify_dataset_exit_code_user_ok(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["user:pan"] = 2
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 0


def test_verify_dataset_exit_code_violation_1(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["setup:flow rate"] = 0
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 3  # zero flow rate and warnings about mismatch


def test_verify_dataset_exit_code_violation_2(monkeypatch):
    # get the exit status from the script
    def sys_exit(status):
        return status

    monkeypatch.setattr(sys, "exit", sys_exit)

    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")

    # Since 0.35.0 we really check the configuration key types.
    # Just make sure that they are properly set:
    with h5py.File(h5path, "a") as h5:
        for key in ["fluorescence:sample rate",
                    "imaging:roi position x",
                    "imaging:roi position y"]:
            h5.attrs[key] = int(h5.attrs[key])

    # provoke a warning
    with h5py.File(h5path, "r+") as h5:
        h5.attrs["setup:flow rate"] = 0
        h5.attrs["setup:flow rate sample"] = 0
        h5.attrs["setup:flow rate sheath"] = 0
    exit_status = cli.verify_dataset(h5path)
    assert exit_status == 2  # zero flow rate



import numpy as np
import dclab
from dclab.external import skimage
from dclab import kde_contours
from dclab import polygon_filter


def test_contour_basic():
    np.random.seed(47)
    x0 = np.random.normal(loc=100, scale=10, size=100)
    y0 = np.random.normal(loc=.1, scale=.01, size=100)

    ds = dclab.new_dataset({"area_um": x0, "deform": y0})
    ds.config["filtering"]["enable filters"] = False

    x, y, kde = ds.get_kde_contour(xax="area_um",
                                   yax="deform",
                                   xacc=.10,
                                   yacc=.01,
                                   kde_type="histogram")
    level = kde_contours.get_quantile_levels(density=kde,
                                             x=x,
                                             y=y,
                                             xp=ds["area_um"],
                                             yp=ds["deform"],
                                             q=.89,
                                             normalize=True)
    contours = kde_contours.find_contours_level(density=kde,
                                                x=x,
                                                y=y,
                                                level=level,
                                                closed=True)

    nump = 0
    for p in zip(x0, y0):
        nump += polygon_filter.PolygonFilter.point_in_poly(p, poly=contours[0])

    assert nump == 11, "there should be (1-q)*100 points in the contour"

    # added in dclab 0.24.1
    nump2 = skimage.measure.points_in_poly(
        np.concatenate((x0.reshape(-1, 1), y0.reshape(-1, 1)), axis=1),
        contours[0])
    assert nump2.sum() == 11


def test_contour_user_put_zero_accuracy():
    np.random.seed(47)
    x0 = np.random.normal(loc=100, scale=10, size=100)
    y0 = np.random.normal(loc=.1, scale=.01, size=100)

    ds = dclab.new_dataset({"area_um": x0, "deform": y0})
    ds.config["filtering"]["enable filters"] = False

    x, y, kde = ds.get_kde_contour(xax="area_um",
                                   yax="deform",
                                   xacc=0,  # testing zero-valued accuracy
                                   yacc=.01,
                                   kde_type="histogram")
    assert np.allclose(x[0][0], 74.24317287410939, atol=1e-12, rtol=0)
    assert np.allclose(y[0][0], 0.07748466975161497, atol=1e-12, rtol=0)
    assert np.allclose(kde[0][0], 0, atol=1e-12, rtol=0)


def test_percentile():
    np.random.seed(47)
    x0 = np.random.normal(loc=100, scale=10, size=1000)
    y0 = np.random.normal(loc=.1, scale=.01, size=1000)

    ds = dclab.new_dataset({"area_um": x0, "deform": y0})
    ds.config["filtering"]["enable filters"] = False

    x, y, kde = ds.get_kde_contour(xax="area_um",
                                   yax="deform",
                                   xacc=.10,
                                   yacc=.01,
                                   kde_type="histogram")
    level = kde_contours.get_quantile_levels(density=kde,
                                             x=x,
                                             y=y,
                                             xp=ds["area_um"],
                                             yp=ds["deform"],
                                             q=.89,
                                             normalize=True)

    level2, err = kde_contours._find_quantile_level(density=kde,
                                                    x=x,
                                                    y=y,
                                                    xp=ds["area_um"],
                                                    yp=ds["deform"],
                                                    quantile=.89,
                                                    acc=0,
                                                    ret_err=True)
    # since _find_quantile level does not do linear interpolation
    # in the density, the computed values can differ from the values
    # obtained using get_quantile_levels - even with err==0.
    assert err == 0
    # This is the resulting level difference.
    assert np.abs(level - level2) < 0.00116



import numbers
import warnings

from dclab.polygon_filter import points_in_poly
import h5py
import matplotlib.pylab as plt
import numpy as np
import scipy.interpolate as spint
from scipy.spatial import ConvexHull
import skimage


from .hooks import lut_hooks


NORM = {
    "area_um": 290,
    "deform": 0.2,
    "volume": 3400,
}


class LutProcessor:
    def __init__(self, hdf5_path, featx="area_um", featy="deform",
                 use_hooks=True, verbose=False):
        """Process simulation data in the HDF5 file format

        The simulation dataset must be an HDF5 file with a specific structure
        that contains the results of FEM simulations. The original HDF5 files
        were provided by Lucas Wittwer. Any LUT added to dclab after version
        0.23.0 was extracted and created using this script.

        The following data post-processing is performed for the LUT:

        - Redundant values in the LUT are removed by checking whether they
          can be reproduced through linear interpolation of the remaining
          values.
        - The LUT is complemented with linearly interpolated isoelasticity
          lines to the convex hull of the simulation data in the area_um
          vs. deform plot.

        Additional post-processing hooks for LUT or isoelastics generation
        are defined in the Python files named according to the LUT identifier
        in the "fem_hooks" subdirectory.

        An example HDF5 file can be found on figshare
        (LE-2D-FEM-19, https://doi.org/10.6084/m9.figshare.12155064.v4).

        Parameters
        ----------
        hdf5_path: str or pathlib.Path
            path to the simulation file
        featx: str
            feature along the x-direction of the LUT
        featy: str
            feature along the y-direction of the LUT
        use_hooks: bool
            whether to perform custom LUT postprocessing using LUT hooks
        verbose: bool
            set to True to display informative plots
        """
        self.verbose = verbose
        self.path = hdf5_path
        self.data, self.meta = self.get_lut_raw()
        self.meta["column features"] = [featx, featy, "emodulus"]
        if use_hooks:
            self.hook = lut_hooks.get(self.meta["identifier"])
            if not self.hook:
                warnings.warn(f"No hook defined for {self.meta['identifier']}")
        else:
            self.hook = None

        # compute the raw LUT as stored in the simulation data
        lutsize = len(self.data["emodulus"])
        lut = np.zeros((lutsize, 3), dtype=float)
        lut[:, 0] = self.data[featx]
        lut[:, 1] = self.data[featy]
        lut[:, 2] = self.data["emodulus"]

        self.lut_raw = lut

        self.featx = featx
        self.featy = featy

        self.xmin = lut[:, 0].min()
        self.xmax = lut[:, 0].max()
        self.xptp = self.xmax - self.xmin

        self.ymin = lut[:, 1].min()
        self.ymax = lut[:, 1].max()
        self.yptp = self.ymax - self.ymin

        # Interpolation grid sizes
        xsize = int(250 * (self.xmax - self.xmin) / NORM[self.featx])
        ysize = int(250 * (self.ymax - self.ymin) / NORM[self.featy])
        self.size = int(np.max([xsize, ysize]))

        # Interpolation grid coordinates
        x = np.linspace(0, 1, self.size, endpoint=True)
        y = np.linspace(0, 1, self.size, endpoint=True)
        self.xitp, self.yitp = np.meshgrid(x, y, indexing="ij")

        # compute the convex hull
        hull = ConvexHull(lut[:, :2])

        self.convex_hull = hull.points[hull.vertices, :]

        if self.hook:
            self.hook.lut_preprocess(self)

    def points_in_convex_hull(self, points):
        convex_hull = np.array(self.convex_hull, copy=True)
        mx, my = np.mean(convex_hull, axis=0)
        for ii in range(len(convex_hull)):
            xi, yi = convex_hull[ii]

            dx = self.xptp / 1000
            if xi < mx:
                dx *= -1
            convex_hull[ii][0] += dx

            dy = self.yptp / 1000
            if yi < my:
                dy *= -1
            convex_hull[ii, 1] += dy
        return points_in_poly(points, convex_hull)

    def normalize_lut(self, lut):
        """Normalize an input LUT to the unit cube"""
        wlut = np.array(lut, copy=True)
        wlut[:, 0] = (wlut[:, 0] - self.xmin) / self.xptp
        wlut[:, 1] = (wlut[:, 1] - self.ymin) / self.yptp
        return wlut

    def denormalize_contour(self, contour):
        """Convert a contour obtained from the pixelated grid to real units"""
        ccu = np.zeros_like(contour)
        ccu[:, 0] = contour[:, 0] / self.size * self.xptp + self.xmin
        ccu[:, 1] = contour[:, 1] / self.size * self.yptp + self.ymin
        return ccu

    def map_lut_to_grid(self, lut=None):
        """Convert LUT data to a 2D image"""
        if lut is None:
            lut = self.lut_raw
        lut = np.array(lut, copy=True)
        lut = lut[lut[:, 0] <= self.xmax, :]
        lut = lut[lut[:, 1] <= self.ymax, :]

        # normalize
        wlut = self.normalize_lut(lut)

        emod = spint.griddata((wlut[:, 0], wlut[:, 1]), wlut[:, 2],
                              (self.xitp, self.yitp), method="linear")

        mask_sim = np.zeros_like(emod, dtype=bool)
        x = self.xitp[:, 0]
        y = self.yitp[0, :]
        for xi, yi, _ in wlut:
            dx = np.abs(x - xi)
            dy = np.abs(y - yi)
            xidx = np.argmin(dx)
            yidx = np.argmin(dy)
            if dx[xidx] + dy[yidx] < 1:
                mask_sim[xidx, yidx] = True

        return emod, mask_sim

    def extract_isoelastics_from_grid(self, emod, lut=None, num=13):
        """Extract isoelasitcs from gridded data

        Parameters
        ----------
        emod: 2d np.ndarray
            array containing Young's modulus gridded according to
            `self.xitp` and `self.yitp`.
        lut: 2d np.ndarray
            LUT (like `self.lut_raw`).
        num: int
            number of isoelastics to compute. They will be evenly
            distributed across the LUT using the coordinates
            `self.xitp` and `self.yitp`. That means that if you pass
            e.g. a smaller LUT, you will not get as many isoealstics
            back, since they will be nan-valued.
        """
        if lut is None:
            lut = self.lut_raw
        wlut = self.normalize_lut(lut)
        # Determine the levels via a line plot through the
        # given LUT.

        # These are the original levels (by Christoph Herold):
        # levels = [0.9 1.2 1.5 1.8 2.1 2.55 3. 3.6 4.2 5.4 6.9]
        # These are the new levels (by the following algorithm for
        # the LE-2D-FEM-19 dataset):
        # levels = [0.91 1.14 1.37 1.63 1.94 2.33 2.83 3.55 4.66 6.64 11.44]
        l0 = np.nanpercentile(emod, 1)
        l1 = np.nanpercentile(emod, 99)
        dl = (l1 - l0) / 200

        blob0 = np.where(np.abs(emod - l0) < .5 * dl)
        p0 = 0, blob0[1][-1]

        blob1 = np.where(np.abs(emod - l1) < .5 * dl)
        p1 = blob1[0][-1], 0

        xlev = np.linspace(self.xitp[p0[0], 0],
                           self.xitp[p1[0], 0], num, endpoint=True)
        ylev = np.linspace(self.yitp[0, p0[1]],
                           self.yitp[0, p1[1]], num, endpoint=True)

        elev = spint.griddata((wlut[:, 0], wlut[:, 1]), wlut[:, 2],
                              (xlev, ylev), method="linear")
        levels = elev[1:-1]

        levels = np.round(levels, 2)

        contours_px = []
        contours = []
        for level in levels:
            conts = skimage.measure.find_contours(emod, level=level)
            if not conts:
                continue
            # get the longest one
            idx = np.argmax([len(cc) for cc in conts])
            cc = conts[idx]
            # remove nan values
            cc = cc[~np.isnan(np.sum(cc, axis=1))]
            # downsample contour on normalized pixel grid
            cc_new = [cc[0]]
            last_used = 0
            for ii in range(1, len(cc)-1):
                if np.sqrt(np.sum((cc[ii] - cc[last_used])**2)) > 1:
                    cc_new.append(cc[ii])
                    last_used = ii
            cc_new.append(cc[-1])
            cc = np.array(cc_new)

            contours_px.append(cc)
            # convert pixel to absolute area_um and deform
            ccu = self.denormalize_contour(cc)
            contours.append(ccu)
        return levels, contours, contours_px

    def get_lut_raw(self):
        """Extract features from a FEM simulation HDF5 file (Lucas Wittwer)

        Returns
        -------
        data: dict
            Each key is a dclab feature, the value is a 1D ndarray
        meta: dict
            Metadata extracted from the HDF5 file
        """
        area_um = []
        deform = []
        emod = []
        volume = []
        with h5py.File(self.path, "r") as h5:
            assert h5.attrs["flow_rate_unit"].decode("utf-8") == "uL/s"
            assert h5.attrs["channel_width_unit"].decode("utf-8") == "um"
            assert h5.attrs["fluid_viscosity_unit"].decode("utf-8") == "Pa s"
            meta = dict(h5.attrs)
            # convert viscosity to mPa*s
            meta["fluid_viscosity_unit"] = "mPa s"
            meta["fluid_viscosity"] = meta["fluid_viscosity"] * 1000
            for key in meta:
                if isinstance(meta[key], bytes):
                    meta[key] = meta[key].decode("utf-8")
                elif isinstance(meta[key], str):
                    pass
                elif isinstance(meta[key], numbers.Integral):
                    meta[key] = int(meta[key])
                else:
                    meta[key] = float(meta[key])
            for Ek in h5.keys():
                assert h5[Ek].attrs["emodulus_unit"].decode("utf-8") == "Pa"
                for simk in h5[Ek].keys():
                    sim = h5[Ek][simk]
                    area_um.append(sim.attrs["area"])
                    assert sim.attrs["area_unit"].decode("utf-8") == "um^2"
                    deform.append(sim.attrs["deformation"])
                    assert sim.attrs["deformation_unit"].decode("utf-8") == ""
                    volume.append(sim.attrs["volume"])
                    assert sim.attrs["volume_unit"].decode("utf-8") == "um^3"
                    emod.append(h5[Ek].attrs["emodulus"] / 1000)
        data = {"area_um": np.array(area_um),
                "deform": np.array(deform),
                "emodulus": np.array(emod),
                "volume": np.array(volume),
                }
        return data, meta

    def compute_isoelastics(self, lut=None, num=13):
        """Compute isoelastics for a give LUT

        The gridded support of the LUT is reduced with a binary_closing
        of size 10.
        """
        if lut is None:
            lut = self.lut_raw

        lut = np.array(lut, copy=True)
        # Map the simulation data to a 2d grid
        emod, _ = self.map_lut_to_grid(lut)

        levels, contours, contours_px = self.extract_isoelastics_from_grid(
            emod=emod, lut=lut, num=num)

        return levels, contours, contours_px

    def assemble_lut_and_isoelastics(self):
        """Apply pipeline, return the LUT, isoelastics, and isoelastics level

        If `self.verbose` is True, then display informative plots.
        """
        emod, mask_sim = self.map_lut_to_grid()
        levels, contours_ip, contours_px = self.compute_isoelastics(num=13)

        if self.hook:
            contours_ip = self.hook.isoelastics_postprocess(self, contours_ip)

        contours = []
        for cc in contours_ip:
            inside = self.points_in_convex_hull(cc[:, :2])
            contours.append(cc[inside])

        if self.verbose:
            plt.figure(figsize=(13, 5))
            ax0 = plt.subplot(131,
                              title="simulation density and mask generation")
            ax0.imshow(mask_sim)

            ax1 = plt.subplot(132, title="isoelastics interpolation on grid")
            ax1.imshow(emod)
            for cont in contours_px:
                ax1.plot(cont[:, 1], cont[:, 0], "x")

            ax2 = plt.subplot(133, title="extracted isoelastics")
            for cont in contours:
                ax2.plot(cont[:, 0], cont[:, 1], "-")
            ax2.set_ylim(0, self.lut_raw[:, 1].max()*1.05)
            ax2.set_xlim(0, self.lut_raw[:, 0].max()*1.05)

            plt.tight_layout()
            plt.show()

        assert len(contours) == len(levels)

        lut = self.shrink_lut()

        return lut, contours, levels

    def shrink_lut(self, lut=None):
        """Remove redundant values in a LUT and enforce the convex hull

        This is achieved by checking whether they can be reproduced
        through linear interpolation from the remaining values.
        """
        print("...Shrinking LUT")
        if lut is None:
            lut = self.lut_raw

        inside = self.points_in_convex_hull(lut[:, :2])
        lut = lut[inside]

        wlut = self.normalize_lut(lut)

        keep = np.ones(wlut.shape[0], dtype=bool)

        checked = np.zeros(wlut.shape[0], dtype=int)

        # Take out a few points and see whether linear interpolation would
        # be enough to recover it.
        for ii in range(wlut.shape[0]):
            if checked[ii] >= 4:
                continue
            ids = np.arange(ii, wlut.shape[0], step=47)
            cur = np.array(keep, copy=True)
            cur[ids] = False
            emod = spint.griddata((wlut[cur, 0], wlut[cur, 1]), wlut[cur, 2],
                                  (wlut[ids, 0], wlut[ids, 1]),
                                  method='linear')

            for em, im in zip(emod, ids):
                checked[im] += 1
                if np.isnan(em):
                    continue
                elif np.allclose(em, wlut[im, 2], atol=1e-5, rtol=1e-4):
                    keep[im] = False
                    checked[im] = 4
        print("... -> Removed {} out of {} data points.".format(
            np.sum(~keep), keep.size))

        lut_new = 1*lut[keep, :]

        return lut_new



"""
This library contains classes and methods for the analysis
of real-time deformability cytometry (RT-DC) datasets.
"""
# flake8: noqa: F401
from . import definitions as dfn
from . import features
from . import isoelastics
from . import kde_contours
from . import kde_methods
from . import lme4
from .polygon_filter import PolygonFilter
from . import rtdc_dataset
from .rtdc_dataset import new_dataset, IntegrityChecker, RTDCWriter
from .rtdc_dataset.feat_temp import (
    register_temporary_feature, set_temporary_feature)
from .rtdc_dataset.feat_anc_ml import (
    MachineLearningFeature, load_modc, load_ml_feature, save_modc)
from .rtdc_dataset.feat_anc_plugin.plugin_feature import (
    PlugInFeature, load_plugin_feature)
from . import statistics

from ._version import __version__, __version_tuple__



import warnings

from ... import features

from .ancillary_feature import AncillaryFeature


def compute_emodulus(mm):
    """Wrapper function for computing the Young's modulus

    Please take a look at the docs :ref:`sec_emodulus_usage`
    for more details on the three cases A, B, and C.
    There are also some sanity checks taking place here.
    """
    calccfg = mm.config["calculation"]

    medium = calccfg.get("emodulus medium", "other").lower()
    temperature = calccfg.get("emodulus temperature", None)
    viscosity = calccfg.get("emodulus viscosity", None)

    if viscosity is not None and medium == "other":
        # sanity checks
        if temperature is not None:
            warnings.warn("The 'emodulus temperature' configuration key is "
                          "ignored if the 'emodulus viscosity' key is set!")
        # Case B from the docs
        return compute_emodulus_visc_only(mm)
    else:
        # sanity checks
        if not isinstance(medium, str):
            raise ValueError(
                f"'emodulus medium' must be a string, got '{medium}'!")
        if medium not in features.emodulus.viscosity.KNOWN_MEDIA:
            raise ValueError(
                f"Only the following media are supported: "
                f"{features.emodulus.viscosity.KNOWN_MEDIA}, got '{medium}'!")
        if viscosity is not None:
            raise ValueError("You must not set the 'emodulus viscosity' "
                             "configuration keyword for known media!")
        # warnings
        if "emodulus viscosity model" not in calccfg:
            warnings.warn("Please specify the 'emodulus viscosity model' "
                          "key in the 'calculation' config segion, falling "
                          "back to 'herold-2017'!",
                          DeprecationWarning)
        # actual function calls
        if temperature is not None:
            # case C from the docs
            temperature = mm.config["calculation"]["emodulus temperature"]
            return compute_emodulus_known_media(mm, temperature=temperature)
        elif "temp" in mm:
            # case A from the docs
            return compute_emodulus_known_media(mm, temperature=mm["temp"])


def compute_emodulus_known_media(mm, temperature):
    """Only use known media and one temperature for all"""
    calccfg = mm.config["calculation"]
    # compute elastic modulus
    emod = features.emodulus.get_emodulus(
        area_um=mm["area_um"],
        deform=mm["deform"],
        medium=calccfg["emodulus medium"],
        channel_width=mm.config["setup"]["channel width"],
        flow_rate=mm.config["setup"]["flow rate"],
        px_um=mm.config["imaging"]["pixel size"],
        temperature=temperature,
        lut_data=calccfg["emodulus lut"],
        visc_model=calccfg.get("emodulus viscosity model", "herold-2017"),
    )
    return emod


def compute_emodulus_visc_only(mm):
    """The user entered the viscosity directly"""
    calccfg = mm.config["calculation"]
    # compute elastic modulus
    emod = features.emodulus.get_emodulus(
        area_um=mm["area_um"],
        deform=mm["deform"],
        medium=calccfg["emodulus viscosity"],
        channel_width=mm.config["setup"]["channel width"],
        flow_rate=mm.config["setup"]["flow rate"],
        px_um=mm.config["imaging"]["pixel size"],
        temperature=None,
        visc_model=None,
        lut_data=calccfg["emodulus lut"],
    )
    return emod


def is_channel(mm):
    """Check whether the measurement was performed in the channel

    If the chip region is not set, then it is assumed to be a
    channel measurement (for backwards compatibility and user-
    friendliness).
    """
    if "setup" in mm.config and "chip region" in mm.config["setup"]:
        region = mm.config["setup"]["chip region"]
        if region == "channel":
            # measured in the channel
            return True
        else:
            # measured in the reservoir
            return False
    else:
        # This might be a testing dictionary or someone who is
        # playing around with data. Avoid disappointments here.
        return True


def register():
    # Please note that registering these things is a delicate business,
    # because the priority has to be chosen carefully.
    # Note that here we have not included the "emodulus viscosity model"
    # configuration keyword. This is checked in the `compute_emodulus`
    # method above and a deprecation warning is issued, so old code
    # does not break immediately.
    for pr, vm in [(1, ["emodulus viscosity model"]),
                   (0, [])  # this is deprecated and should be removed!
                   ]:
        AncillaryFeature(feature_name="emodulus",
                         method=compute_emodulus,
                         data="case C",
                         req_features=["area_um", "deform"],
                         req_config=[["calculation", vm + [
                                        "emodulus lut",
                                        "emodulus medium",
                                        "emodulus temperature"]],
                                     ["imaging", ["pixel size"]],
                                     ["setup", ["flow rate", "channel width"]]
                                     ],
                         req_func=is_channel,
                         priority=4 + pr)
        AncillaryFeature(feature_name="emodulus",
                         data="case A",
                         method=compute_emodulus,
                         req_features=["area_um", "deform", "temp"],
                         req_config=[["calculation", vm + [
                                        "emodulus lut",
                                        "emodulus medium"]],
                                     ["imaging", ["pixel size"]],
                                     ["setup", ["flow rate", "channel width"]]
                                     ],
                         req_func=is_channel,
                         priority=0 + pr)

    AncillaryFeature(feature_name="emodulus",
                     data="case B",
                     method=compute_emodulus,
                     req_features=["area_um", "deform"],
                     req_config=[["calculation", vm + [
                                    "emodulus lut",
                                    "emodulus viscosity"]],
                                 ["imaging", ["pixel size"]],
                                 ["setup", ["flow rate", "channel width"]]
                                 ],
                     req_func=is_channel,
                     priority=2)



"""Computation of event contour from event mask"""
from collections import deque
import numbers

import numpy as np

# equivalent to
# from skimage.measure import find_contours
from ..external.skimage.measure import find_contours


class NoValidContourFoundError(BaseException):
    pass


class LazyContourList(object):
    def __init__(self, masks, max_events=1000):
        """A list-like object that computes contours upon indexing

        Parameters
        ----------
        masks: array-like
            3D array of masks, may be an HDF5 dataset or any other
            structure that supports indexing
        max_events: int
            maximum number of contours to keep in the contour list;
            set to 0/False/None to cache all contours

        .. versionchanged:: 0.58.3

            Added the `max_events` parameter which now makes this class
            a lazy, least-recently-used contour list. To achieve the
            old behavior (which may fill up your memory), set
            `max_events=None`.
        """
        self.masks = masks
        self.contours = deque(maxlen=max_events or None)
        self.indices = deque(maxlen=max_events or None)
        #: used for hashing in ancillary features
        self.identifier = str(masks[0][:].tobytes())
        self.shape = len(masks), np.nan, 2

    def __getitem__(self, idx):
        """Compute contour(s) if not already in self.contours"""
        if not isinstance(idx, numbers.Integral):
            # slicing!
            indices = np.arange(len(self))[idx]
            output = []
            # populate the output list
            for evid in indices:
                output.append(self.__getitem__(evid))
            return output
        else:
            try:
                # Is the contour already available?
                idx_q = self.indices.index(idx)
            except ValueError:
                # The contour is not there. Compute it.
                try:
                    cont = get_contour(self.masks[idx])
                except BaseException as e:
                    e.args = (f"Event {idx}, {e.args[0]}",)
                    raise
            else:
                # Get the contour from deque
                cont = self.contours[idx_q]
            # If we got here, it means that we have computed a contour
            # successfully. Store it in the deque.
            self.contours.append(cont)
            self.indices.append(idx)
            return cont

    def __len__(self):
        return len(self.masks)


def get_contour(mask):
    """Compute the image contour from a mask

    The contour is computed in a very inefficient way using scikit-image
    and a conversion of float coordinates to pixel coordinates.

    Parameters
    ----------
    mask: binary ndarray of shape (M,N) or (K,M,N)
        The mask outlining the pixel positions of the event.
        If a 3d array is given, then `K` indexes the individual
        contours.

    Returns
    -------
    cont: ndarray or list of K ndarrays of shape (J,2)
        A 2D array that holds the contour of an event (in pixels)
        e.g. obtained using `mm.contour` where  `mm` is an instance
        of `RTDCBase`. The first and second columns of `cont`
        correspond to the x- and y-coordinates of the contour.
    """
    if isinstance(mask, np.ndarray) and len(mask.shape) == 2:
        mask = [mask]
        ret_list = False
    else:
        ret_list = True
    contours = []

    for mi in mask:
        # This is only 10% slower than doing:
        # conts, _ = cv2.findContours(np.array(mi, dtype=np.uint8),
        #                             cv2.RETR_EXTERNAL,
        #                             cv2.CHAIN_APPROX_NONE)
        # c2 = conts[0].reshape(-1, 2)
        conts = find_contours(mi.transpose(),
                              level=.9999,
                              positive_orientation="low",
                              fully_connected="high")
        # get the longest contour
        c0 = sorted(conts, key=lambda x: len(x))[-1]
        # round all coordinates to pixel values
        c1 = np.asarray(np.round(c0), int)
        # remove duplicates
        c2 = remove_duplicates(c1)
        if len(c2) == 0:
            raise NoValidContourFoundError("No contour found!")
        contours.append(c2)
    if ret_list:
        return contours
    else:
        return contours[0]


def get_contour_lazily(mask):
    """Like :func:`get_contour`, but computes contours on demand

    Parameters
    ----------
    mask: binary ndarray of shape (M,N) or (K,M,N)
        The mask outlining the pixel positions of the event.
        If a 3d array is given, then `K` indexes the individual
        contours.

    Returns
    -------
    cont: ndarray or LazyContourList of K ndarrays of shape (J,2)
        A 2D array that holds the contour of an event (in pixels)
        e.g. obtained using `mm.contour` where  `mm` is an instance
        of `RTDCBase`. The first and second columns of `cont`
        correspond to the x- and y-coordinates of the contour.
    """
    if isinstance(mask, np.ndarray) and len(mask.shape) == 2:
        # same behavior as `get_contour`
        cont = get_contour(mask=mask)
    else:
        cont = LazyContourList(masks=mask)
    return cont


def remove_duplicates(cont):
    """Remove duplicates in a circular contour"""
    x = np.resize(cont, (len(cont) + 1, 2))
    selection = np.ones(len(x), dtype=bool)
    selection[1:] = ~np.prod((x[1:] == x[:-1]), axis=1, dtype=bool)
    return x[selection][:-1]



"""RT-DC hdf5 format"""
from __future__ import annotations

import pathlib

from .. import feat_basin

from .base import RTDC_HDF5


class HDF5Basin(feat_basin.Basin):
    basin_format = "hdf5"
    basin_type = "file"

    def __init__(self, *args, **kwargs):
        self._available_verified = None
        super(HDF5Basin, self).__init__(*args, **kwargs)

    def _load_dataset(self, location, **kwargs):
        return RTDC_HDF5(location, **kwargs)

    def is_available(self):
        if self._available_verified is None:
            with self._av_check_lock:
                try:
                    self._available_verified = \
                        pathlib.Path(self.location).exists()
                except OSError:
                    pass
        return self._available_verified



import hashlib
import io
import json
from unittest import mock
import sys
import time

import dclab
from dclab import cli, new_dataset, rtdc_dataset, RTDCWriter

import h5py
import numpy as np
import pytest

from helper_methods import retrieve_data


def test_check_suffix_disabled_compress():
    path_in_o = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_in = path_in_o.with_suffix("")
    path_in_o.rename(path_in)
    assert path_in.suffix == ""
    with pytest.raises(ValueError, match="Unsupported file type"):
        cli.compress(path_in=path_in,
                     path_out=path_in.with_name("compressed.rtdc"))
    # but this should work:
    cli.compress(path_in=path_in,
                 path_out=path_in.with_name("compressed2.rtdc"),
                 check_suffix=False)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("compressed.rtdc")

    ret = cli.compress(path_in=path_in, path_out=path_out)
    assert ret is None, "by default, this method should return 0 (exit 0)"
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "dclab-compress" in dsj.logs
        assert len(dsj)
        assert len(dsj) == len(ds0)
        for feat in ds0.features:
            if feat in ["contour", "image", "mask"]:
                for ii in range(len(dsj)):
                    assert np.all(dsj[feat][ii] == ds0[feat][ii]), feat
            else:
                assert np.all(dsj[feat] == ds0[feat]), feat


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress_wo_logs():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    with h5py.File(path_in, "a") as h5:
        del h5["logs"]
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("compressed.rtdc")

    cli.compress(path_in=path_in, path_out=path_out)
    with new_dataset(path_out) as ds:
        assert len(ds.logs) == 1


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress_already_compressed():
    """By default, an already compressed dataset should not be compressed"""
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out1 = path_in.with_name("compressed_1.rtdc")
    path_out2 = path_in.with_name("compressed_2.rtdc")
    path_out3 = path_in.with_name("compressed_copy_of_1.rtdc")
    # this is straight-forward
    cli.compress(path_out=path_out1, path_in=path_in)
    # just for the sake of comparison
    time.sleep(1)  # we need different time stamps in path_out2
    cli.compress(path_out=path_out2, path_in=path_in)
    # this is not trivial
    cli.compress(path_out=path_out3, path_in=path_out1)

    # the first two files should not be the same (dates are written, etc)
    h1 = hashlib.md5(path_out1.read_bytes()).hexdigest()
    h2 = hashlib.md5(path_out2.read_bytes()).hexdigest()
    h3 = hashlib.md5(path_out3.read_bytes()).hexdigest()
    assert h1 != h2
    # Changed in dclab 0.49.0: Since the compression step should also check
    # for defective features, it is important to revisit the entire file.
    # As such, it is cleaner to rewrite the entire dataset, since we can now
    # copy single HDF5 Datasets without having to redo the compression.
    # assert h1 == h3
    assert h1 != h3


def test_compress_basin_internal():
    """
    Internal basins should just be copied to the new file
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("compressed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data={"userdef1": np.arange(2)},
                       basin_map=np.zeros(src["events/deform"].shape[0]),
                       basin_feats=["userdef1"],
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "userdef1" in ds.features_basin
        assert "userdef1" not in ds.features_innate

    # compress the basin-based dataset
    cli.compress(path_in=h5path_small, path_out=h5path_out)

    with h5py.File(h5path_out) as h5:
        assert "deform" in h5["events"], "sanity check"
        assert "userdef1" not in h5["events"]
        assert "userdef1" in h5["basin_events"]
        assert np.all(h5["basin_events"]["userdef1"] == np.arange(2))

    with new_dataset(h5path_out) as ds:
        assert "userdef1" in ds.features_basin


def test_compress_basin_no_data_from_basins():
    """
    When compressing a dataset, feature data from the basin should not be
    written to the output file.
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("compressed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "image" in ds.features

    # compress the basin-based dataset
    cli.compress(path_in=h5path_small, path_out=h5path_out)

    with h5py.File(h5path_out) as h5:
        assert "deform" in h5["events"], "sanity check"
        assert "image" not in h5["events"], "Arrgh, basin feature was copied"

    with new_dataset(h5path_out) as ds:
        assert "image" in ds.features_basin


def test_compress_basin_preserved_compress():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Empty, basin-based dataset creation
    with RTDCWriter(h5path_small) as hw:
        bn_hash = hw.store_basin(basin_name="example basin",
                                 basin_type="file",
                                 basin_format="hdf5",
                                 basin_locs=[h5path],
                                 basin_descr="an example test basin",
                                 )
    # compress the data
    h5path_out = h5path_small.with_name("compressed.rtdc")
    cli.compress(path_in=h5path_small,
                 path_out=h5path_out,
                 )

    with h5py.File(h5path_out) as h5_out, h5py.File(h5path_small) as h5_in:
        # check if h5path_out is empty
        if 'events' in h5_out:
            assert len(h5_out['events']) == 0
        # check if h5path_small basin is same as h5path_out basin
        assert 'basins' in h5_in
        assert 'basins' in h5_out
        assert np.all(
            h5_in['basins'][bn_hash][:] == h5_out['basins'][bn_hash][:]
        )

    with new_dataset(h5path) as ds, new_dataset(h5path_out) as ds_out:
        # check if all features of h5path are preserved in h5path_out
        assert len(ds.features_innate) == 32
        assert len(ds_out.features_innate) == 0
        for feat in ds.features_innate:
            assert feat in ds_out.features_basin


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress_log_md5_5m():
    """In dclab 0.42.0 we changed sha256 to md5-5M file checksums"""
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # sanity check (file is < 5MB)
    h1 = hashlib.md5(path_in.read_bytes()).hexdigest()
    assert h1 == "e49db02274ac75ab24911f893c41f5b0"
    # same directory (will be cleaned up with path_in)
    path_out1 = path_in.with_name("compressed_1.rtdc")
    cli.compress(path_out=path_out1, path_in=path_in)
    with dclab.new_dataset(path_out1) as ds:
        log = ds.logs["dclab-compress"]
    dcdict = json.loads("\n".join(log))
    file = dcdict["files"][0]
    assert file["index"] == 1
    assert file["name"] == "mask_contour_reference.rtdc"
    assert file["md5-5M"] == "e49db02274ac75ab24911f893c41f5b0"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_compress_with_online_polygon_filters():
    """Shape-In 2.3 supports online polygon filters"""
    path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # add an artificial online polygon filter
    with h5py.File(path, "a") as h5:
        # set soft filter to True
        h5.attrs["online_filter:area_um,deform soft limit"] = True
        # set filter values
        pf_name = "online_filter:area_um,deform polygon points"
        area_um = h5["events"]["area_um"]
        deform = h5["events"]["deform"]
        pf_points = np.array([
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform)],
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform) + np.std(deform)],
            [np.mean(area_um),
             np.mean(deform) + np.std(deform)],
        ])
        h5.attrs[pf_name] = pf_points

    path_out = path.with_name("compressed.rtdc")
    cli.compress(path_out=path_out, path_in=path)

    with dclab.new_dataset(path_out) as ds:
        assert len(ds) == 8
        assert ds.config["online_filter"]["area_um,deform soft limit"]
        assert "area_um,deform polygon points" in ds.config["online_filter"]
        assert np.allclose(
            ds.config["online_filter"]["area_um,deform polygon points"],
            pf_points)


def test_compress_with_online_polygon_filters_real_data():
    """Shape-In 2.3 supports online polygon filters"""
    path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")

    path_out = path.with_name("compressed.rtdc")
    cli.compress(path_out=path_out, path_in=path)

    with dclab.new_dataset(path_out) as ds:
        assert len(ds) == 1
        assert ds.config["online_filter"]["size_x,size_y soft limit"]
        assert "size_x,size_y polygon points" in ds.config["online_filter"]
        assert np.allclose(
            ds.config["online_filter"]["size_x,size_y polygon points"],
            [[0.1, 0.2],
             [0.1, 2.5],
             [3.3, 3.2],
             [5.2, 0.9]]
        )


@mock.patch('sys.stdout', new_callable=io.StringIO)
def test_version(mock_stdout, monkeypatch):
    def sys_exit(status):
        return status
    monkeypatch.setattr(sys, "exit", sys_exit)

    monkeypatch.setattr(sys, "argv", ["dclab-compress", "--version"])

    parser = cli.compress_parser()
    parser.parse_args()

    stdout_printed = mock_stdout.getvalue()
    assert stdout_printed.count("dclab-compress")
    assert stdout_printed.count(dclab.__version__)



"""Load RT-DC datasets"""
import errno
import io
import os
import pathlib

from .core import RTDCBase
from . import (
    fmt_dict, fmt_dcor, fmt_hdf5, fmt_hierarchy, fmt_http, fmt_s3, fmt_tdms)


def load_file(path, identifier=None, **kwargs):
    path = pathlib.Path(path).resolve()
    for fmt in [fmt_hdf5.RTDC_HDF5, fmt_tdms.RTDC_TDMS]:
        if fmt.can_open(path):
            return fmt(path, identifier=identifier, **kwargs)
    else:
        raise ValueError("Unknown file format: '{}'".format(path.suffix))


def new_dataset(data, identifier=None, **kwargs):
    """Initialize a new RT-DC dataset

    Parameters
    ----------
    data:
        can be one of the following:

        - dict
        - .tdms file
        - .rtdc file
        - subclass of `RTDCBase`
          (will create a hierarchy child)
        - DCOR resource URL
        - URL to file in S3-compatible object store
    identifier: str
        A unique identifier for this dataset. If set to `None`
        an identifier is generated.
    kwargs:
        Additional parameters passed to the RTDCBase subclass

    Returns
    -------
    dataset: subclass of :class:`dclab.rtdc_dataset.RTDCBase`
        A new dataset instance
    """
    if isinstance(data, dict):
        return fmt_dict.RTDC_Dict(data, identifier=identifier, **kwargs)
    elif isinstance(data, io.BytesIO):
        return fmt_hdf5.RTDC_HDF5(data, **kwargs)
    elif fmt_dcor.is_dcor_url(data):
        return fmt_dcor.RTDC_DCOR(data, identifier=identifier, **kwargs)
    elif fmt_http.is_http_url(data):
        if fmt_http.is_url_available(data, ret_reason=False):
            return fmt_http.RTDC_HTTP(data, identifier=identifier)
        elif fmt_s3.is_s3_url(data):
            return fmt_s3.RTDC_S3(data, identifier=identifier, **kwargs)
        else:
            raise NotImplementedError(f"Unknown remote format: {data}")
    elif isinstance(data, RTDCBase):
        return fmt_hierarchy.RTDC_Hierarchy(data, identifier=identifier,
                                            **kwargs)
    elif isinstance(data, (pathlib.Path, str)):
        # If we are given a `file:` specifier, remove it. We might be
        # looking at a network share (not mounted as a drive) on Windows
        # (e.g. "file:\\127.0.0.1\shared\data.rtdc").
        if isinstance(data, str) and data.startswith("file:"):
            data = data[5:]
        path = pathlib.Path(data).resolve()
        if not path.exists():
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), str(path))
        else:
            return load_file(data, identifier=identifier, **kwargs)
    else:
        msg = "data type not supported: {}".format(data.__class__)
        raise NotImplementedError(msg)





from ... import features
from .ancillary_feature import AncillaryFeature


class MissingCrosstalkMatrixElementsError(BaseException):
    pass


def compute_ctc(mm, fl_channel):
    if "fl1_max" in mm:
        fl1 = mm["fl1_max"]
    else:
        fl1 = 0

    if "fl2_max" in mm:
        fl2 = mm["fl2_max"]
    else:
        fl2 = 0

    if "fl3_max" in mm:
        fl3 = mm["fl3_max"]
    else:
        fl3 = 0

    ctdict = {}

    for i in [1, 2, 3]:
        for j in [1, 2, 3]:
            if i == j:
                continue
            key = "crosstalk fl{}{}".format(i, j)
            par = "ct{}{}".format(i, j)
            if key in mm.config["calculation"]:
                ctdict[par] = mm.config["calculation"][key]

    if ("fl1_max" in mm and
        "fl2_max" in mm and
        "fl3_max" in mm and
        ("ct12" not in ctdict or
         "ct13" not in ctdict or
         "ct21" not in ctdict or
         "ct23" not in ctdict or
         "ct31" not in ctdict or
         "ct32" not in ctdict)):
        msg = "{}, has fl1_max, fl2_max, and fl3_max,".format(mm) \
              + " but not all crosstalk matrix elements are" \
              + " defined in the 'calculation' configuration section."
        raise MissingCrosstalkMatrixElementsError(msg)

    return features.fl_crosstalk.correct_crosstalk(
        fl1=fl1,
        fl2=fl2,
        fl3=fl3,
        fl_channel=fl_channel,
        **ctdict)


def compute_ctc1(mm):
    return compute_ctc(mm, fl_channel=1)


def compute_ctc2(mm):
    return compute_ctc(mm, fl_channel=2)


def compute_ctc3(mm):
    return compute_ctc(mm, fl_channel=3)


def get_method(fl_channel):
    if fl_channel == 1:
        return compute_ctc1
    elif fl_channel == 2:
        return compute_ctc2
    elif fl_channel == 3:
        return compute_ctc3


def register():
    opts_all = (["fl1_max",
                 "fl2_max",
                 "fl3_max"],
                ["crosstalk fl21",
                 "crosstalk fl31",
                 "crosstalk fl12",
                 "crosstalk fl32",
                 "crosstalk fl13",
                 "crosstalk fl23"])

    opts_12 = (["fl1_max",
                "fl2_max"],
               ["crosstalk fl21",
                "crosstalk fl12"])

    opts_13 = (["fl1_max",
                "fl3_max"],
               ["crosstalk fl31",
                "crosstalk fl13"])

    opts_23 = (["fl2_max",
                "fl3_max"],
               ["crosstalk fl32",
                "crosstalk fl23"])

    for flch in [1, 2, 3]:
        AncillaryFeature(feature_name="fl{}_max_ctc".format(flch),
                         method=get_method(flch),
                         req_features=opts_all[0],
                         req_config=[["calculation", opts_all[1]]],
                         priority=1)

    for flch in [1, 2]:
        AncillaryFeature(feature_name="fl{}_max_ctc".format(flch),
                         method=get_method(flch),
                         req_features=opts_12[0],
                         req_config=[["calculation", opts_12[1]]],
                         priority=0)

    for flch in [1, 3]:
        AncillaryFeature(feature_name="fl{}_max_ctc".format(flch),
                         method=get_method(flch),
                         req_features=opts_13[0],
                         req_config=[["calculation", opts_13[1]]],
                         priority=0)

    for flch in [2, 3]:
        AncillaryFeature(feature_name="fl{}_max_ctc".format(flch),
                         method=get_method(flch),
                         req_features=opts_23[0],
                         req_config=[["calculation", opts_23[1]]],
                         priority=0)



"""Helper methods for copying .rtdc data"""
from __future__ import annotations

import json
import re
from typing import List, Literal

import h5py
import h5py.h5o
import hdf5plugin
import numpy as np

from ..definitions import feature_exists, scalar_feature_exists
from ..util import hashobj

from .fmt_hdf5 import DEFECTIVE_FEATURES, RTDC_HDF5
from .writer import RTDCWriter


def rtdc_copy(src_h5file: h5py.Group,
              dst_h5file: h5py.Group,
              features: List[str] | Literal['all', 'scalar', 'none'] = "all",
              include_basins: bool = True,
              include_logs: bool = True,
              include_tables: bool = True,
              meta_prefix: str = ""):
    """Create a compressed copy of an RT-DC file

    Parameters
    ----------
    src_h5file: h5py.Group
        Input HDF5 file
    dst_h5file: h5py.Group
        Output HDF5 file
    features: list of strings or one of ['all', 'scalar', 'none']
        If this is a list then it specifies the features that are copied from
        `src_h5file` to `dst_h5file`. Alternatively, you may specify 'all'
        (copy all features), 'scalar' (copy only scalar features), or 'none'
        (don't copy any features).
    include_basins: bool
        Copy the basin information from `src_h5file` to `dst_h5file`.
    include_logs: bool
        Copy the logs from `src_h5file` to `dst_h5file`.
    include_tables: bool
        Copy the tables from `src_h5file` to `dst_h5file`.
    meta_prefix: str
        Add this prefix to the name of the logs and tables in `dst_h5file`.
    """
    # metadata
    dst_h5file.attrs.update(src_h5file.attrs)

    # events in source file
    if "events" in src_h5file:
        events_src = list(src_h5file["events"].keys())
    else:
        events_src = []

    if include_basins and "basin_events" in src_h5file:
        events_src += list(src_h5file["basin_events"].keys())
        events_src = sorted(set(events_src))

    # logs
    if include_logs and "logs" in src_h5file:
        dst_h5file.require_group("logs")
        for l_key in src_h5file["logs"]:
            h5ds_copy(src_loc=src_h5file["logs"],
                      src_name=l_key,
                      dst_loc=dst_h5file["logs"],
                      dst_name=meta_prefix + l_key,
                      recursive=False)

    # tables
    if include_tables and "tables" in src_h5file:
        dst_h5file.require_group("tables")
        for tkey in src_h5file["tables"]:
            # There appears to be a problem with h5copy in some rare
            # situations, so we do not use h5copy, but read and write
            # the table data directly.
            # https://github.com/HDFGroup/hdf5/issues/3214
            # The following caused a Segmentation fault:
            # h5ds_copy(src_loc=src_h5file["tables"],
            #           src_name=tkey,
            #           dst_loc=dst_h5file["tables"],
            #           dst_name=meta_prefix + tkey,
            #           recursive=False)
            copy_table = dst_h5file["tables"].create_dataset(
                name=tkey,
                data=src_h5file["tables"][tkey][:],
                fletcher32=True,
                **hdf5plugin.Zstd(clevel=5))
            copy_table.attrs.update(src_h5file["tables"][tkey].attrs)

    # events
    if isinstance(features, list):
        feature_iter = features
    elif features == "all":
        feature_iter = events_src
    elif features == "scalar":
        feature_iter = [feat for feat in events_src
                        if feature_exists(feat, scalar_only=True)]
    elif features == "none":
        feature_iter = []
    else:
        raise ValueError(f"`features` must be either a list of feature names "
                         f"or one of 'all', 'scalar' or 'none', got "
                         f"'{features}'")

    # Additional check for basin features.
    bn_regexp = re.compile("^basinmap[0-9]*$")  # future-proof regexp
    src_basin_feats = [f for f in events_src if bn_regexp.match(f)]
    if include_basins:
        # Make sure all 'basinmap?' features are included in the output file.
        for feat in src_basin_feats:
            if feat not in feature_iter:
                feature_iter.append(feat)
    else:
        # We do not need the basinmap features, because basins are
        # stripped from the output file.
        for feat in src_basin_feats:
            if feat in feature_iter:
                feature_iter.remove(feat)

    # copy basin definitions
    if include_basins and "basins" in src_h5file:
        basin_definition_copy(src_h5file=src_h5file,
                              dst_h5file=dst_h5file,
                              features_iter=feature_iter)

    if feature_iter:
        dst_h5file.require_group("events")
        for feat in feature_iter:
            if not feature_exists(feat):
                continue
            elif feat in src_h5file["events"]:
                # Skip all defective features. These are features that
                # are known to be invalid (e.g. ancillary features that
                # were computed falsely) and must be recomputed by dclab.
                if feat in DEFECTIVE_FEATURES:
                    defective = DEFECTIVE_FEATURES[feat](src_h5file)
                    if defective:
                        continue

                dst = h5ds_copy(src_loc=src_h5file["events"],
                                src_name=feat,
                                dst_loc=dst_h5file["events"],
                                recursive=True)
                if scalar_feature_exists(feat):
                    # complement min/max values for all scalar features
                    for ufunc, attr in [(np.nanmin, "min"),
                                        (np.nanmax, "max"),
                                        (np.nanmean, "mean"),
                                        ]:
                        if attr not in dst.attrs:
                            dst.attrs[attr] = ufunc(dst)

            elif (include_basins
                    and "basin_events" in src_h5file
                    and feat in src_h5file["basin_events"]):
                # Also copy internal basins which should have been defined
                # in the "basin_events" group.
                if feat in src_h5file["basin_events"]:
                    h5ds_copy(src_loc=src_h5file["basin_events"],
                              src_name=feat,
                              dst_loc=dst_h5file.require_group("basin_events"),
                              dst_name=feat
                              )


def basin_definition_copy(src_h5file, dst_h5file, features_iter):
    """Copy basin definitions `src_h5file["basins"]` to the new file

    Normally, we would just use :func:`h5ds_copy` to copy basins from
    one dataset to another. However, if we are e.g. only copying scalar
    features, and there are non-scalar features in the internal basin,
    then we must rewrite the basin definition of the internal basin.

    The `features_iter` list of features defines which features are
    relevant for the internal basin.
    """
    dst_h5file.require_group("basins")
    # Load the basin information
    basin_dicts = RTDC_HDF5.basin_get_dicts_from_h5file(src_h5file)
    for bn in basin_dicts:
        b_key = bn["key"]

        if b_key in dst_h5file["basins"]:
            # already stored therein
            continue

        # sanity check
        if b_key not in src_h5file["basins"]:
            raise ValueError(
                f"Failed to parse basin information correctly. Source file "
                f"{src_h5file} does not contain basin {b_key} which I got "
                f"from `RTDC_HDF5.basin_get_dicts_from_h5file`.")

        if bn["type"] == "internal":
            # Make sure we define the internal features selected
            feat_used = [f for f in bn["features"] if f in features_iter]
            if len(feat_used) == 0:
                # We don't have any internal features, don't write anything
                continue
            elif feat_used != bn["features"]:
                bn["features"] = feat_used
                rewrite = True
            else:
                rewrite = False
        else:
            # We do not have an internal basin, just copy everything
            rewrite = False

        if rewrite:
            # Convert edited `bn` to JSON and write feature data
            b_lines = json.dumps(bn, indent=2).split("\n")
            key = hashobj(b_lines)
            if key not in dst_h5file["basins"]:
                with RTDCWriter(dst_h5file) as hw:
                    hw.write_text(dst_h5file["basins"], key, b_lines)
        else:
            # copy only
            h5ds_copy(src_loc=src_h5file["basins"],
                      src_name=b_key,
                      dst_loc=dst_h5file["basins"],
                      dst_name=b_key,
                      recursive=False)


def h5ds_copy(src_loc, src_name, dst_loc, dst_name=None,
              ensure_compression=True, recursive=True):
    """Copy an HDF5 Dataset from one group to another

    Parameters
    ----------
    src_loc: h5py.H5Group
        The source location
    src_name: str
        Name of the dataset in `src_loc`
    dst_loc: h5py.H5Group
        The destination location
    dst_name: str
        The name of the destination dataset, defaults to `src_name`
    ensure_compression: bool
        Whether to make sure that the data are compressed,
        If disabled, then all data from the source will be
        just copied and not compressed.
    recursive: bool
        Whether to recurse into HDF5 Groups (this is required e.g.
        for copying the "trace" feature)

    Returns
    -------
    dst: h5py.Dataset
        The dataset `dst_loc[dst_name]`

    Raises
    ------
    ValueError:
        If the named source is not a h5py.Dataset
    """
    compression_kwargs = hdf5plugin.Zstd(clevel=5)
    dst_name = dst_name or src_name
    src = src_loc[src_name]
    if isinstance(src, h5py.Dataset):
        if ensure_compression and not is_properly_compressed(src):
            # Chunk size larger than dataset size is not allowed
            # in h5py's `make_new_dset`.
            if src.shape[0] == 0:
                # Ignore empty datasets (This sometimes happens with logs).
                return
            elif src.chunks and src.chunks[0] > src.shape[0]:
                # The chunks in the input file are larger than the dataset
                # shape. So we set the chunks to the shape. Here, we only
                # check for the first axis (event count for feature data),
                # because if the chunks vary in any other dimension then
                # there is something fundamentally wrong with the input
                # dataset (which we don't want to endorse, and where there
                # could potentially be a lot of data put into ram).
                chunks = list(src.chunks)
                chunks[0] = src.shape[0]
                chunks = tuple(chunks)
            else:
                # original chunk size is fine
                chunks = src.chunks
            # Variable length strings, compression, and fletcher32 are not
            # a good combination. If we encounter any logs, then we have
            # to write them with fixed-length strings.
            # https://forum.hdfgroup.org/t/fletcher32-filter-on-variable-
            # length-string-datasets-not-suitable-for-filters/9038/4
            if src.dtype.kind == "O":
                # We are looking at logs with variable length strings.
                max_length = max([len(ii) for ii in src] + [100])
                dtype = f"S{max_length}"
                convert_to_s_fixed = True
            else:
                dtype = src.dtype
                convert_to_s_fixed = False

            # Manually create a compressed version of the dataset.
            dst = dst_loc.create_dataset(name=dst_name,
                                         shape=src.shape,
                                         dtype=dtype,
                                         chunks=chunks,
                                         fletcher32=True,
                                         **compression_kwargs
                                         )
            if convert_to_s_fixed:
                # We are looking at old variable-length log strings.
                dst[:] = src[:].astype(dtype)
            elif chunks is None:
                dst[:] = src[:]
            else:
                for chunk in src.iter_chunks():
                    dst[chunk] = src[chunk]
            # Also write all the attributes
            dst.attrs.update(src.attrs)
        else:
            # Copy the Dataset to the destination as-is.
            h5py.h5o.copy(src_loc=src_loc.id,
                          src_name=src_name.encode(),
                          dst_loc=dst_loc.id,
                          dst_name=dst_name.encode(),
                          )
    elif recursive and isinstance(src, h5py.Group):
        dst_rec = dst_loc.require_group(dst_name)
        for key in src:
            h5ds_copy(src_loc=src,
                      src_name=key,
                      dst_loc=dst_rec,
                      ensure_compression=ensure_compression,
                      recursive=recursive)
    else:
        raise ValueError(f"The object {src_name} in {src.file} is not "
                         f"a dataset!")
    return dst_loc[dst_name]


def is_properly_compressed(h5obj):
    """Check whether an HDF5 object is properly compressed

    The compression check only returns True if the input file was
    compressed with the Zstandard compression using compression
    level 5 or higher.
    """
    # Since version 0.43.0, we use Zstandard compression
    # which does not show up in the `compression`
    # attribute of `obj`.
    create_plist = h5obj.id.get_create_plist()
    filter_args = create_plist.get_filter_by_id(32015)
    if filter_args is not None and filter_args[1][0] >= 5:
        properly_compressed = True
    else:
        properly_compressed = False
    return properly_compressed



# flake8: noqa: F401
"""DCOR client interface"""
from .api import REQUESTS_AVAILABLE
from .base import (
    DCOR_CERTS_SEARCH_PATHS, RTDC_DCOR, get_server_cert_path, is_dcor_url
)
from .basin import DCORBasin



import numbers

from . import feat_logic, meta_const, meta_parse


def config_key_exists(section, key):
    """Return `True` if the configuration key exists"""
    valid = False
    if section == "user":
        if isinstance(key, str) and key.strip():  # sanity check
            valid = True
    elif meta_const.config_funcs.get(section, {}).get(key, False):
        valid = True
    elif section == "online_filter":
        if (key.count(",")
                and (key.endswith("soft limit")
                     or key.endswith("polygon points"))):
            # "online_filter:area_um,deform soft limit"
            # "online_filter:area_um,deform polygon points"
            f1, f2 = key.split(" ", 1)[0].split(",")
            valid = (feat_logic.scalar_feature_exists(f1)
                     and feat_logic.scalar_feature_exists(f2))
        else:
            feat = key.split(" ", 1)[0]
            valid = feat_logic.scalar_feature_exists(feat)
    return valid


def get_config_value_descr(section, key):
    """Return the description of a config value

    Returns `key` if not defined anywhere
    """
    descr = key
    if section == "user":
        pass
    elif meta_const.config_descr.get(section, {}).get(key, False):
        descr = meta_const.config_descr[section][key]
    elif section == "online_filter":
        if (key.count(",")
                and (key.endswith("soft limit")
                     or key.endswith("polygon points"))):
            # "online_filter:area_um,deform soft limit"
            # "online_filter:area_um,deform polygon points"
            f1, f2 = key.split(" ", 1)[0].split(",")
            # do not include units
            l1 = feat_logic.get_feature_label(f1, with_unit=False)
            l2 = feat_logic.get_feature_label(f2, with_unit=False)
            if key.endswith("soft limit"):
                descr = f"Soft limit, polygon ({l1}, {l2})"
            elif key.endswith("polygon points"):
                descr = f"Polygon ({l1}, {l2})"
        else:
            feat = key.split(" ", 1)[0]
            if feat_logic.scalar_feature_exists(feat):
                if key.endswith("soft limit"):
                    # "online_filter:area_um soft limit"
                    label = feat_logic.get_feature_label(feat, with_unit=False)
                    descr = f"Soft limit, {label}"
                elif key.endswith("min"):
                    # "online_filter:area_um min"
                    descr = f"Min. {feat_logic.get_feature_label(feat)}"
                elif key.endswith("max"):
                    # "online_filter:area_um max"
                    descr = f"Max. {feat_logic.get_feature_label(feat)}"

    return descr


def get_config_value_func(section, key):
    """Return configuration type converter function"""
    func = None
    if section == "user":
        pass
    elif meta_const.config_funcs.get(section, {}).get(key, False):
        func = meta_const.config_funcs[section][key]
    elif section == "online_filter":
        # Note that for "min" and "max" values we do nothing (None)
        if key.endswith("soft limit"):
            # "online_filter:area_um,deform soft limit"
            func = meta_parse.fbool
        elif key.endswith("polygon points"):
            # "online_filter:area_um,deform polygon points"
            func = meta_parse.f2dfloatarray

    if func is None:
        return lambda x: x
    else:
        return func


def get_config_value_type(section, key):
    """Return the expected type of a config value

    Returns `None` if no type is defined
    """
    typ = None
    if section == "user":
        pass
    elif meta_const.config_types.get(section, {}).get(key, False):
        typ = meta_const.config_types[section][key]
    elif section == "online_filter":
        if key.endswith("soft limit"):
            # "online_filter:area_um,deform soft limit"
            typ = meta_parse.func_types[meta_parse.fbool]
        elif key.endswith("polygon points"):
            typ = meta_parse.func_types[meta_parse.f2dfloatarray]
        elif key.endswith("min") or key.endswith("max"):
            # most-general type is a number
            typ = numbers.Number
    return typ



"""Create .rtdc files with scalar-only features"""
from __future__ import annotations

import argparse
import pathlib
from typing import List
import warnings

import h5py
import hdf5plugin

from ..rtdc_dataset import (
    fmt_hdf5, new_dataset, rtdc_copy, RTDCWriter, RTDCBase
)
from .. import util
from .._version import version

from . import common


def condense(
        path_in: str | pathlib.Path = None,
        path_out: str | pathlib.Path = None,
        ancillaries: bool = None,
        store_ancillary_features: bool = True,
        store_basin_features: bool = True,
        check_suffix: bool = True,
        ret_path: bool = False
        ):
    """Create a new dataset with all available scalar-only features

    Besides the innate scalar features, this also includes all
    fast-to-compute ancillary and all basin features (`features_loaded`).

    Parameters
    ----------
    path_in: str or pathlib.Path
        file to compress
    path_out: str or pathlib
        output file path
    ancillaries: bool
        DEPRECATED, use `store_ancillary_features` instead
    store_ancillary_features: bool
        compute and store ancillary features in the output file
    store_basin_features: bool
        copy basin features from the input path to the output file;
        Note that the basin information (including any internal
        basin dataset) are always copied over to the new dataset.
    check_suffix: bool
        check suffixes for input and output files
    ret_path: bool
        whether to return the output path

    Returns
    -------
    path_out: pathlib.Path (optional)
        output path (with possibly corrected suffix)
    """
    if ancillaries is not None:
        warnings.warn("Please use `store_ancillary_features` instead of "
                      "`ancillaries`", DeprecationWarning)
        store_ancillary_features = ancillaries

    if path_out is None or path_in is None:
        parser = condense_parser()
        args = parser.parse_args()
        path_in = args.input
        path_out = args.output
        store_ancillary_features = not args.no_ancillaries
        store_basin_features = not args.no_basins

    allowed_input_suffixes = [".rtdc", ".tdms"]
    if not check_suffix:
        allowed_input_suffixes.append(pathlib.Path(path_in).suffix)

    path_in, path_out, path_temp = common.setup_task_paths(
        path_in, path_out, allowed_input_suffixes=allowed_input_suffixes)

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        # We use `store_basin_features` during initialization (to avoid
        # conflicts with ancillary features) and in the actual function
        # as well, to correctly determine which features to use.
        with new_dataset(path_in, enable_basins=store_basin_features) as ds, \
                h5py.File(path_temp, "w") as h5_cond:
            condense_dataset(ds=ds,
                             h5_cond=h5_cond,
                             store_ancillary_features=store_ancillary_features,
                             store_basin_features=store_basin_features,
                             warnings_list=w)

    # Finally, rename temp to out
    path_temp.rename(path_out)
    if ret_path:
        return path_out


def condense_dataset(
        ds: RTDCBase,
        h5_cond: h5py.File,
        ancillaries: bool = None,
        store_ancillary_features: bool = True,
        store_basin_features: bool = True,
        warnings_list: List = None):
    """Condense a dataset using low-level HDF5 methods

    For ancillary and basin features, high-level dclab methods are used.
    """
    if ancillaries is not None:
        warnings.warn("Please use `store_ancillary_features` instead of "
                      "`ancillaries`", DeprecationWarning)
        store_ancillary_features = ancillaries

    cmp_kw = hdf5plugin.Zstd(clevel=5)
    cmd_dict = {}

    # If we have an input HDF5 file, then we might readily copy most
    # of the features over using rtdc_copy. If we have a .tdms file,
    # then we have to go the long route.
    if isinstance(ds, fmt_hdf5.RTDC_HDF5):
        rtdc_copy(src_h5file=ds.h5file,
                  dst_h5file=h5_cond,
                  features="scalar",
                  include_basins=True,
                  include_logs=True,
                  include_tables=True,
                  meta_prefix="")

    h5_cond.require_group("logs")

    # scalar features
    feats_sc = ds.features_scalar
    # loaded (computationally cheap) scalar features
    feats_sc_loaded = [f for f in ds.features_loaded if f in feats_sc]
    # internal basin features that have already been copied with `rtdc_copy`
    feats_sc_basint = sorted(h5_cond.get("basin_events", {}).keys())
    # features that are excluded, because we already copied them
    feats_exclude = feats_sc_loaded + feats_sc_basint

    cmd_dict["features_original_innate"] = ds.features_innate

    features = set(feats_sc_loaded)
    if store_basin_features:
        feats_sc_basin = [f for f in ds.features_basin if
                          (f in feats_sc and f not in feats_exclude)]
        cmd_dict["features_basin"] = feats_sc_basin
        if feats_sc_basin:
            print(f"Using basin features {feats_sc_basin}")
            features |= set(feats_sc_basin)

    if store_ancillary_features:
        feats_sc_anc = [f for f in ds.features_ancillary if
                        (f in feats_sc and f not in feats_exclude)]
        cmd_dict["features_ancillary"] = feats_sc_anc
        if feats_sc_anc:
            features |= set(feats_sc_anc)
            print(f"Using ancillary features {feats_sc_anc}")

    # command log
    logs = {"dclab-condense": common.get_command_log(
        paths=[ds.path], custom_dict=cmd_dict)}

    # rename old dclab-condense logs
    for l_key in ["dclab-condense", "dclab-condense-warnings"]:
        if l_key in h5_cond["logs"]:
            # This is cached, so no worry calling it multiple times.
            md5_cfg = util.hashobj(ds.config)
            # rename
            new_log_name = f"{l_key}_{md5_cfg}"
            if new_log_name not in h5_cond["logs"]:
                # If the user repeatedly condensed one file, then there is
                # no benefit in storing the log under a different name (the
                # metadata did not change). Only write the log if it does
                # not already exist.
                h5_cond["logs"][f"{l_key}_{md5_cfg}"] = h5_cond["logs"][l_key]
            del h5_cond["logs"][l_key]

    with RTDCWriter(h5_cond,
                    mode="append",
                    compression_kwargs=cmp_kw,
                    ) as hw:
        # Write all remaining scalar features to the file
        # (these are *all* scalar features in the case of .tdms data).
        for feat in features:
            if feat not in h5_cond["events"]:
                hw.store_feature(feat=feat, data=ds[feat])

        # collect warnings log
        if warnings_list:
            logs["dclab-condense-warnings"] = \
                common.assemble_warnings(warnings_list)

        # Write logs
        for name in logs:
            hw.store_log(name, logs[name])


def condense_parser():
    descr = "Reduce an RT-DC measurement to its scalar-only features " \
            + "(i.e. without `contour`, `image`, `mask`, or `trace`). " \
            + "All available ancillary features are computed."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('input', metavar="INPUT", type=str,
                        help='Input path (.tdms or .rtdc file)')
    parser.add_argument('output', metavar="OUTPUT", type=str,
                        help='Output path (.rtdc file)')
    parser.add_argument('--no-ancillary-features',
                        dest='no_ancillaries',
                        action='store_true',
                        help='Do not compute expensive ancillary features '
                             'such as volume'
                        )
    parser.set_defaults(no_ancillaries=False)
    parser.add_argument('--no-basin-features',
                        dest='no_basins',
                        action='store_true',
                        help='Do not store basin-based feature data from the '
                             'input file in the output file'
                        )
    parser.set_defaults(no_basins=False)
    parser.add_argument('--version', action='version',
                        version=f'dclab-condense {version}')
    return parser



"""R lme4 wrapper"""
import logging
import numbers
import pathlib
import tempfile

import importlib_resources
import numpy as np

from .. import definitions as dfn
from ..rtdc_dataset.core import RTDCBase

from . import rsetup


logger = logging.getLogger(__name__)


class Rlme4(object):
    def __init__(self, model="lmer", feature="deform"):
        """Perform an R-lme4 analysis with RT-DC data

        Parameters
        ----------
        model: str
            One of:

            - "lmer": linear mixed model using lme4's ``lmer``
            - "glmer+loglink": generalized linear mixed model using
              lme4's ``glmer`` with an additional a log-link function
              via the ``family=Gamma(link='log'))`` keyword.
        feature: str
            Dclab feature for which to compute the model
        """
        #: modeling method to use (e.g. "lmer")
        self.model = None
        #: dclab feature for which to perform the analysis
        self.feature = None
        #: list of [RTDCBase, column, repetition, chip_region]
        self.data = []

        self.set_options(model=model, feature=feature)

        # Make sure that lme4 is available
        if not rsetup.has_lme4():
            logger.info("Installing lme4, this may take a while!")
            rsetup.require_lme4()

    def add_dataset(self, ds, group, repetition):
        """Add a dataset to the analysis list

        Parameters
        ----------
        ds: RTDCBase
            Dataset
        group: str
            The group the measurement belongs to ("control" or
            "treatment")
        repetition: int
            Repetition of the measurement

        Notes
        -----
        - For each repetition, there must be a "treatment" (``1``) and a
          "control" (``0``) group.
        - If you would like to perform a differential feature analysis,
          then you need to pass at least a reservoir and a channel
          dataset (with same parameters for `group` and `repetition`).
        """
        assert group in ["treatment", "control"]
        assert isinstance(ds, RTDCBase)
        assert isinstance(repetition, numbers.Integral)

        region = ds.config["setup"]["chip region"]
        # make sure there are no doublets
        for ii, dd in enumerate(self.data):
            if dd[1] == group and dd[2] == repetition and dd[3] == region:
                raise ValueError("A dataset with group '{}', ".format(group)
                                 + "repetition '{}', and ".format(repetition)
                                 + "'{}' region has already ".format(region)
                                 + "been added (index {})!".format(ii))

        self.data.append([ds, group, repetition, region])

    def check_data(self):
        """Perform sanity checks on ``self.data``"""
        # Check that we have enough data
        if len(self.data) < 3:
            msg = "Linear mixed effects models require repeated " \
                  + "measurements. Please add more repetitions."
            raise ValueError(msg)

    def fit(self, model=None, feature=None):
        """Perform (generalized) linear mixed-effects model fit

        The response variable is modeled using two linear mixed effect
        models:

        - model: "feature ~ group + (1 + group | repetition)"
          (random intercept + random slope model)
        - the null model: "feature ~ (1 + group | repetition)"
          (without the fixed effect introduced by the "treatment" group).

        Both models are compared in R using "anova" (from the
        R-package "stats" :cite:`Everitt1992`) which performs a
        likelihood ratio test to obtain the p-Value for the
        significance of the fixed effect (treatment).

        If the input datasets contain data from the "reservoir"
        region, then the analysis is performed for the differential
        feature.

        Parameters
        ----------
        model: str (optional)
            One of:

            - "lmer": linear mixed model using lme4's ``lmer``
            - "glmer+loglink": generalized linear mixed model using
              lme4's ``glmer`` with an additional log-link function
              via ``family=Gamma(link='log'))`` :cite:`lme4`
        feature: str (optional)
            dclab feature for which to compute the model

        Returns
        -------
        results: dict
            Dictionary with the results of the fitting process:

            - "anova p-value": Anova likelihood ratio test (significance)
            - "feature": name of the feature used for the analysis
              ``self.feature``
            - "fixed effects intercept": Mean of ``self.feature`` for all
              controls; In the case of the "glmer+loglink" model, the intercept
              is already back transformed from log space.
            - "fixed effects treatment": The fixed effect size between the mean
              of the controls and the mean of the treatments relative to
              "fixed effects intercept"; In the case of the "glmer+loglink"
              model, the fixed effect is already back transformed from log
              space.
            - "fixed effects repetitions": The effects (intercept and
              treatment) for each repetition. The first axis defines
              intercept/treatment; the second axis enumerates the repetitions;
              thus the shape is (2, number of repetitions) and
              ``np.mean(results["fixed effects repetitions"], axis=1)`` is
              equivalent to the tuple (``results["fixed effects intercept"]``,
              ``results["fixed effects treatment"]``) for the "lmer" model.
              This does not hold for the "glmer+loglink" model, because
              of the non-linear inverse transform back from log space.
            - "is differential": Boolean indicating whether or not
              the analysis was performed for the differential (bootstrapped
              and subtracted reservoir from channel data) feature
            - "model": model name used for the analysis ``self.model``
            - "model converged": boolean indicating whether the model
              converged
            - "r model summary": Summary of the model
            - "r model coefficients": Model coefficient table
            - "r script": the R script used
            - "r output": full output of the R script
        """
        self.set_options(model=model, feature=feature)
        self.check_data()

        # Assemble dataset
        if self.is_differential():
            # bootstrap and compute differential features using reservoir
            features, groups, repetitions = self.get_differential_dataset()
        else:
            # regular feature analysis
            features = []
            groups = []
            repetitions = []
            for dd in self.data:
                features.append(self.get_feature_data(dd[1], dd[2]))
                groups.append(dd[1])
                repetitions.append(dd[2])

        # concatenate and populate arrays for R
        features_c = np.concatenate(features)
        groups_c = np.zeros(len(features_c), dtype=str)
        repetitions_c = np.zeros(len(features_c), dtype=int)
        pos = 0
        for ii in range(len(features)):
            size = len(features[ii])
            groups_c[pos:pos+size] = groups[ii][0]
            repetitions_c[pos:pos+size] = repetitions[ii]
            pos += size

        # Run R with the given template script
        rscript = importlib_resources.read_text("dclab.lme4",
                                                "lme4_template.R")
        _, script_path = tempfile.mkstemp(prefix="dclab_lme4_", suffix=".R",
                                          text=True)
        script_path = pathlib.Path(script_path)
        rscript = rscript.replace("<MODEL_NAME>", self.model)
        rscript = rscript.replace("<FEATURES>", arr2str(features_c))
        rscript = rscript.replace("<REPETITIONS>", arr2str(repetitions_c))
        rscript = rscript.replace("<GROUPS>", arr2str(groups_c))
        script_path.write_text(rscript, encoding="utf-8")

        result = rsetup.run_command((rsetup.get_r_script_path(), script_path))

        ret_dict = self.parse_result(result)
        ret_dict["is differential"] = self.is_differential()
        ret_dict["feature"] = self.feature
        ret_dict["r script"] = rscript
        ret_dict["r output"] = result
        assert ret_dict["model"] == self.model

        return ret_dict

    def get_differential_dataset(self):
        """Return the differential dataset for channel/reservoir data

        The most famous use case is differential deformation. The idea
        is that you cannot tell what the difference in deformation
        from channel to reservoir, because you never measure the
        same object in the reservoir and the channel. You usually just
        have two distributions. Comparing distributions is possible
        via bootstrapping. And then, instead of running the lme4
        analysis with the channel deformation data, it is run with
        the differential deformation (subtraction of the bootstrapped
        deformation distributions for channel and reservoir).
        """
        features = []
        groups = []
        repetitions = []
        # compute differential features
        for grp in sorted(set([dd[1] for dd in self.data])):
            # repetitions per groups
            grp_rep = sorted(set([dd[2] for dd in self.data if dd[1] == grp]))
            for rep in grp_rep:
                feat_cha = self.get_feature_data(grp, rep, region="channel")
                feat_res = self.get_feature_data(grp, rep, region="reservoir")
                bs_cha, bs_res = bootstrapped_median_distributions(feat_cha,
                                                                   feat_res)
                # differential feature
                features.append(bs_cha - bs_res)
                groups.append(grp)
                repetitions.append(rep)
        return features, groups, repetitions

    def get_feature_data(self, group, repetition, region="channel"):
        """Return array containing feature data

        Parameters
        ----------
        group: str
            Measurement group ("control" or "treatment")
        repetition: int
            Measurement repetition
        region: str
            Either "channel" or "reservoir"

        Returns
        -------
        fdata: 1d ndarray
            Feature data (Nans and Infs removed)
        """
        assert group in ["control", "treatment"]
        assert isinstance(repetition, numbers.Integral)
        assert region in ["reservoir", "channel"]
        for dd in self.data:
            if dd[1] == group and dd[2] == repetition and dd[3] == region:
                ds = dd[0]
                break
        else:
            raise ValueError("Dataset for group '{}', repetition".format(group)
                             + " '{}', and region".format(repetition)
                             + " '{}' not found!".format(region))
        fdata = ds[self.feature][ds.filter.all]
        fdata_valid = fdata[~np.logical_or(np.isnan(fdata), np.isinf(fdata))]
        return fdata_valid

    def is_differential(self):
        """Return True if the differential feature is computed for analysis

        This effectively just checks the regions of the datasets
        and returns True if any one of the regions is "reservoir".

        See Also
        --------
        get_differential_features: for an explanation
        """
        for dd in self.data:
            if dd[3] == "reservoir":
                return True
        else:
            return False

    def parse_result(self, result):
        resd = result.split("OUTPUT")
        ret_dict = {}
        for item in resd:
            string = item.split("#*#")[0]
            key, value = string.split(":", 1)
            key = key.strip()
            value = value.strip().replace("\n\n", "\n")

            if key == "fixed effects repetitions":
                rows = value.split("\n")[1:]
                reps = []
                for row in rows:
                    reps.append([float(vv) for vv in row.split()[1:]])
                value = np.array(reps).transpose()
            elif key == "model converged":
                value = value == "TRUE"
            elif value == "NA":
                value = np.nan
            else:
                try:
                    value = float(value)
                except ValueError:
                    pass

            ret_dict[key] = value
        return ret_dict

    def set_options(self, model=None, feature=None):
        """Set analysis options"""
        if model is not None:
            assert model in ["lmer", "glmer+loglink"]
            self.model = model
        if feature is not None:
            assert dfn.scalar_feature_exists(feature)
            self.feature = feature


def arr2str(a):
    """Convert an array to a string"""
    if isinstance(a.dtype.type, np.integer):
        return ",".join(str(dd) for dd in a.tolist())
    elif a.dtype.type == np.str_:
        return ",".join(f"'{dd}'" for dd in a.tolist())
    else:
        return ",".join(f"{dd:.16g}" for dd in a.tolist())


def bootstrapped_median_distributions(a, b, bs_iter=1000, rs=117):
    """Compute the bootstrapped distributions for two arrays.

    Parameters
    ----------
    a, b: 1d ndarray of length N
        Input data
    bs_iter: int
        Number of bootstrapping iterations to perform
        (output size).
    rs: int
        Random state seed for random number generator

    Returns
    -------
    median_dist_a, median_dist_b: 1d arrays of length bs_iter
        Boostrap distribution of medians for ``a`` and ``b``.

    See Also
    --------
    `<https://en.wikipedia.org/wiki/Bootstrapping_(statistics)>`_

    Notes
    -----
    From a programmatic point of view, it would have been better
    to implement this method for just one input array (because of
    redundant code). However, due to historical reasons (testing
    and comparability to Shape-Out 1), bootstrapping is done
    interleaved for the two arrays.
    """
    # Seed random numbers that are reproducible on different machines
    prng_object = np.random.RandomState(rs)
    # Initialize median arrays
    median_a = np.zeros(bs_iter)
    median_b = np.zeros(bs_iter)
    # If this loop is still too slow, we could get rid of it and
    # do everything with arrays. Depends on whether we will
    # eventually run into memory problems with array sizes
    # of y*bs_iter and yR*bs_iter.
    lena = len(a)
    lenb = len(b)
    for q in range(bs_iter):
        # Compute random indices and draw from a, b
        draw_a_idx = prng_object.randint(0, lena, lena)
        median_a[q] = np.median(a[draw_a_idx])
        draw_b_idx = prng_object.randint(0, lenb, lenb)
        median_b[q] = np.median(b[draw_b_idx])
    return median_a, median_b



"""Basic methods for event feature computation"""
from . import contour  # noqa: F401
from . import bright  # noqa: F401
from . import bright_bc  # noqa: F401
from . import bright_perc  # noqa: F401
from . import emodulus  # noqa: F401
from . import fl_crosstalk  # noqa: F401
from . import inert_ratio  # noqa: F401
from . import volume  # noqa: F401



"""Pixelation correction for emodulus look-up-tables and isoelastics

Use this script to study and visualize how pixelation affects
individual features and especially how it affects look-up-tables
for emodulus computation and isoelastics.

This script partially reproduces figures 2, 4, and 5 of Christoph
Herold's famous arXiv manuscript (https://arxiv.org/abs/1704.00572).

As an argument, you should pass a FEM simulation results HDF5 file
(https://doi.org/10.6084/m9.figshare.12155064.v4), which is required
for obtaining a volume-deformation look-up table and for visualizing
pixelation effects using in-silico data (run fem2rtdc.py first).

Please also see dclab issue #70 for a broader view on things.

Notes
-----

I added two example output .png files to this directory:

 - pixelation_correction_2020.png: This is the original output file from
   when this script was first implemented.
 - pixelation_correction_2022.png: This is a revised version from 2022
   which properly implements random offsets in the disk's mask images
   (see dclab issue #178).
"""
import argparse
import os
import pathlib

import dclab
import lmfit
import matplotlib.pylab as plt
import numpy as np


from fem2rtdc import features_from_mask
from lut_recipes import LutProcessor

#: You can edit the pixel size to see how the pixelation effect scales
#: with the pixel size. The fits incorporate this parameter to always
#: yield the same results, but the reference implementations also
#: do this. Not that this does not affect the black dots, which are
#: only plotted for comparison and were generated with fem2rtdc.py.
PIXEL_SIZE = 0.34  # um/px


def get_lut_volume(path):
    lup = LutProcessor(path, use_hooks=True, featx="volume")
    lut, _, _ = lup.assemble_lut_and_isoelastics()
    return lut, lup.meta


def create_multiexp_model(n=5, x=None, x0=None, y=None, decaycoeffs=None,
                          offset=None):
    """Create a multiexponential model with smart initial values for lmfit"""
    modelexpr = ["off"]
    for ii in range(n):
        modelexpr.append("A{} * exp(-(x-x0)/t{})".format(ii, ii))

    mod = lmfit.models.ExpressionModel(" + ".join(modelexpr))
    params = mod.make_params()

    if x is not None:
        params["x0"].set(value=x.min(), min=-np.abs(x).max()/20, max=np.min(x))

    params["off"].set(min=0, max=2, value=.1)

    if x0 is not None:
        params["x0"].set(value=x0, vary=False)

    amplitudes = [.01] * n

    if decaycoeffs is None:
        vary_tau = True
        if x is not None:
            decaycoeffs = np.logspace(-n, 0, n, base=3)/5*(x.max()-x.min())
        else:
            decaycoeffs = np.logspace(0, 4, n)

    else:
        assert len(decaycoeffs) == n
        vary_tau = False

    for ii in range(n):
        if x is not None:
            tmax = x.max() - x.min()
        else:
            tmax = 10**n
        params["t{}".format(ii)].set(min=1e-5, max=tmax,
                                     value=decaycoeffs[ii], vary=vary_tau)
        if y is not None:
            amax = 2*(y.max() - y.min())
        else:
            amax = 1

        params["A{}".format(ii)].set(min=1e-5, max=amax, value=amplitudes[ii])

    if offset is None and y is not None:
        idoff = x > x.max()*.9
        params["off"].set(value=np.mean(y[idoff]))
    else:
        params["off"].set(value=offset, vary=False)

    return mod, params


def disk_image(area_um, random_displacement=True, seed=None):
    """Compute pixelated disk image for a given event area

    The disk is mapped on a grid. A pixel is assumed to be
    inside the disk when its center point is within the
    perimeter of the disk.

    Parameters
    ----------
    area_um: float
        Accurate area [µm²] of the event
    random_displacement: bool
        If set to True, randomly displace the center of the
        disk by up to 1 px in x and y. If set to False, the center of
        the disk is at the center between two pixels.
    seed: int or None
        Random seed to use when moving the disk in the x-y-plane
        before performing the mapping onto the grid. Set to None if
        you have to set the seed someplace else.

    Returns
    -------
    mask: 2D boolean ndarray
        Binary disk image
    """
    # radius in pixels
    radius = np.sqrt(area_um / np.pi) / PIXEL_SIZE

    size = int(np.ceil(radius*2.2))
    if size % 2:
        size += 1
    center = size / 2

    # determine suitable mesh
    x = np.arange(size)
    xm, ym = np.meshgrid(x, x, indexing="ij")

    if not random_displacement:
        # centered between pixels
        offx = 0.5
        offy = 0.5
    else:
        # random offsets
        if seed is not None:
            np.random.seed(seed)
        offx = np.random.uniform(0, 1)
        offy = np.random.uniform(0, 1)
    xm = xm - center + offx
    ym = ym - center + offy

    # binary disk image
    mask = (xm**2 + ym**2) <= radius**2

    return mask


def fit_exponential(dataf, datax, decaycoeffs, x0=None, offset=None):
    """Fit a multi-exponential decay

    This is used for reproducing the pixelation correction.

    Parameters
    ----------
    dataf: 1D ndarray of length N
        The y-data to fit
    datax: 1D ndarray of length N
        The corresponding x data
    decaycoeffs: list of float
        The decay times (coefficients). They are defined
        in `create_multiexp_model`. They are not varied
        during fitting.
    x0: float or None
        If None, then the constant x-offset is a fit parameter.
        If a float, then the constant x-offset is fixed.
    offset: float or None
        If None, then the constant y-offset is a fit parameter
        (initial value determined from `dataf` for large `datax`).
        If a float, then the constant y-offset is fixed.

    Returns
    -------
    result: lmfit fitting results object
    """
    mod, params = create_multiexp_model(n=len(decaycoeffs),
                                        x=datax,
                                        x0=x0,
                                        offset=offset,
                                        decaycoeffs=decaycoeffs,
                                        y=dataf)

    result = mod.fit(dataf, params=params, x=datax)
    return result


def fit_exponential_free(dataf, datax, n, x0=None, offset=None):
    """Fit a multi-exponential decay (prototyping)

    This is used for finding the decay times (coefficients)
    for `fit_exponential`. The initial values of the decay
    times are spaced logarithmically. You can increase `n` and
    then just remove those coefficients that did not vary during
    fitting.

    Parameters
    ----------
    dataf: 1D ndarray of length N
        The y-data to fit
    datax: 1D ndarray of length N
        The corresponding x data
    n: int
        Number of exponentials to fit.
    x0: float or None
        If None, then the constant x-offset is a fit parameter.
        If a float, then the constant x-offset is fixed.
    offset: float or None
        If None, then the constant y-offset is a fit parameter
        (initial value determined from `dataf` for large `datax`).
        If a float, then the constant y-offset is fixed.

    Returns
    -------
    result: lmfit fitting results object
    """
    mod, params = create_multiexp_model(n=n,
                                        x=datax,
                                        offset=offset,
                                        x0=x0,
                                        y=dataf)

    result = mod.fit(dataf, params=params, x=datax)
    return result


def corr_deform_with_area_um(area_um, px_um=PIXEL_SIZE):
    """original from Christoph Herold"""
    pxscale = (.34 / px_um)**2
    offs = 0.0012
    exp1 = 0.020 * np.exp(-area_um * pxscale / 7.1)
    exp2 = 0.010 * np.exp(-area_um * pxscale / 38.6)
    exp3 = 0.005 * np.exp(-area_um * pxscale / 296)
    delta = offs + exp1 + exp2 + exp3
    return delta


def corr_area_um_with_area_um(area_um, px_um=PIXEL_SIZE):
    """from here"""
    pxscale = (.34 / px_um)**2
    offs = 1.006
    exp1 = 0.625 * np.exp(-area_um * pxscale / 3.5)
    exp2 = 0.106 * np.exp(-area_um * pxscale / 20)
    exp3 = 0.041 * np.exp(-area_um * pxscale / 86)
    exp4 = 0.021 * np.exp(-area_um * pxscale / 463)
    delta = offs + exp1 + exp2 + exp3 + exp4
    return delta


def corr_deform_with_volume(volume, px_um=PIXEL_SIZE):
    """from here"""
    pxscalev = (.34 / px_um)**3
    offs = 0.0013
    exp1 = 0.0172 * np.exp(-volume * pxscalev / 40)
    exp2 = 0.0070 * np.exp(-volume * pxscalev / 450)
    exp3 = 0.0032 * np.exp(-volume * pxscalev / 6040)
    delta = offs + exp1 + exp2 + exp3
    return delta


def corr_volume_with_volume(volume, px_um=PIXEL_SIZE):
    """from here"""
    pxscalev = (.34 / px_um)**3
    offs = 1.023
    exp1 = 0.719 * np.exp(-volume * pxscalev / 8.3)
    exp2 = 0.123 * np.exp(-volume * pxscalev / 102)
    exp3 = 0.061 * np.exp(-volume * pxscalev / 885)
    exp4 = 0.035 * np.exp(-volume * pxscalev / 8830)
    delta = offs + exp1 + exp2 + exp3 + exp4
    return delta


def plot_numerical_data(path, ax1, ax2, ax3, ax4):
    """Load and plot numerical data exported using femt2rtdc.py

    Parameters
    ----------
    path: pathlib.Path
        Path to a simulation file (this file is actually
        not used, only the output of fem2rtdc.py).
    ax1: matplotlib.Axis
        deform error versus area_um plot
    ax2: matplotlib.Axis
        area_um error versus area_um plot
    ax3: matplotlib.Axis
        deform error versus volume plot
    ax4: matplotlib.Axis
        volume error versus volume plot
    """
    h5mapped = sorted(path.parent.glob("*.map*.rtdc"))
    h5orig = path.with_suffix(".orig.rtdc")

    dso = dclab.new_dataset(h5orig)
    dsm = [dclab.new_dataset(pp) for pp in h5mapped]

    area_um = np.concatenate([ds["area_um"] for ds in dsm])
    deform = np.concatenate([ds["deform"] for ds in dsm])
    volume = np.concatenate([ds["volume"] for ds in dsm])

    true_area_um = np.concatenate([dso["area_um"]]*5)
    true_deform = np.concatenate([dso["deform"]]*5)
    true_volume = np.concatenate([dso["volume"]]*5)

    delta_deform = deform - true_deform
    delta_area_um = true_area_um / area_um
    delta_volume = true_volume / volume

    plotkw = dict(ms=1, color="k", alpha=.05, label="FEM simulation")

    ax1.plot(area_um, delta_deform, ".", **plotkw)
    ax2.plot(area_um, delta_area_um, ".", **plotkw)
    ax3.plot(volume, delta_deform, ".", **plotkw)
    ax4.plot(volume, delta_volume, ".", **plotkw)


def plot_and_fit_disk_data(ax1, ax2, ax3, ax4):
    """Visualize the pixelation effect for deform, area, and volume

    Reproduces the mostly figures 2 and 4 of Christoph Herold's
    famous arXiv paper and goes beyond that.

    We are only investigating discs/spheres here, because Herold
    already showed that using an ellipse does not really change
    things. In addition, we can overlay the disk plot with in-silico
    data from the FEM dataset (see `plot_numerical_data`).

    Parameters
    ----------
    ax1: matplotlib.Axis
        deform error versus area_um plot
    ax2: matplotlib.Axis
        area_um error versus area_um plot
    ax3: matplotlib.Axis
        deform error versus volume plot
    ax4: matplotlib.Axis
        volume error versus volume plot

    Notes
    -----
    - Computed data are cached as .npz in the script directory.
    - The fitting region goes much beyond the region shown in
      the plot produced by this script (__main__ section).
    - The reference fit for deform-vs-area_um plot is taken
      from Herold's manuscript (this is why it does not fully
      coincide with the fit here).
    """
    # sanity check (circle center positioned at pixel corner)
    assert disk_image(np.pi*(PIXEL_SIZE*2)**2, False).sum() == 12

    N = 10000
    R = 5

    here = pathlib.Path(__file__).parent
    path = here / ".cache_pixelation_correction_{}.npz".format(PIXEL_SIZE)
    if not os.path.exists(path):
        area_um = np.zeros(N*R, dtype=float)
        deform = np.zeros(N*R, dtype=float)
        volume = np.zeros(N*R, dtype=float)

        true_area_um = np.zeros(N*R, dtype=float)
        true_deform = np.zeros(N*R, dtype=float)
        true_volume = np.zeros(N*R, dtype=float)

        for rr in np.arange(R):
            np.random.seed(R)
            for ii, ari in enumerate(np.linspace(10, 1250, N)):
                # Shape-In values
                mask = disk_image(ari)
                feats = features_from_mask(mask, pixel_size=PIXEL_SIZE)
                area_um[rr + ii*R] = feats["area_um"]
                deform[rr + ii*R] = feats["deform"]
                volume[rr + ii*R] = feats["volume"]
                # True values
                true_area_um[rr + ii*R] = ari
                radius = np.sqrt(ari/np.pi)
                true_volume[rr + ii*R] = 4/3 * np.pi * radius**3
                true_deform[rr + ii*R] = 0  # circle

        delta_deform = deform - true_deform
        delta_area_um = true_area_um / area_um
        delta_volume = true_volume / volume

        np.savez(path,
                 delta_deform=delta_deform,
                 delta_area_um=delta_area_um,
                 delta_volume=delta_volume,
                 area_um=area_um,
                 deform=deform,
                 volume=volume,
                 )
    else:
        dd = np.load(path)
        delta_deform = dd["delta_deform"]
        delta_area_um = dd["delta_area_um"]
        delta_volume = dd["delta_volume"]
        area_um = dd["area_um"]
        deform = dd["deform"]
        volume = dd["volume"]

    # Pixelation effects scale with pixel size
    pxscale = (.34 / PIXEL_SIZE)**2
    pxscalev = (.34 / PIXEL_SIZE)**3

    # Figure 2 Herold: deform-area_um
    ax1.set_title("Deviation of deformation (area_um)")
    ax1.set_ylabel("deform - true_deform")

    aref = np.linspace(area_um.min(), area_um.max(), 1000)
    # Fit with fixed a0, b0, c0
    result_deform = fit_exponential(dataf=delta_deform,
                                    datax=area_um * pxscale,
                                    x0=0,
                                    offset=0.0012,
                                    decaycoeffs=[7.1, 38.6, 296])

    # Plot comparison
    ax1.plot(area_um, delta_deform, ".", ms=1, label="sphere")
    ax1.plot(area_um, result_deform.best_fit, label="fit to sphere data")
    ax1.plot(aref, corr_deform_with_area_um(aref), linestyle=(
        0, (5, 10)), color="k", alpha=.5, label="reference fit")

    # Figure 4 (only circle) Herold: area_um-area_um
    ax2.set_title("Deviation of area (area_um)")
    ax2.set_ylabel("true_area_um / area_um")

    result_area_um = fit_exponential(dataf=delta_area_um,
                                     datax=area_um * pxscale,
                                     x0=0,
                                     offset=1.006,
                                     decaycoeffs=[3.5, 20, 86, 463])

    ax2.plot(area_um, delta_area_um, ".", ms=1, label="sphere")
    ax2.plot(area_um, result_area_um.best_fit, label="fit to sphere data")
    ax2.plot(aref, corr_area_um_with_area_um(aref), linestyle=(
        0, (5, 10)), color="k", alpha=.5, label="reference fit")

    # New figure deform-volume
    ax3.set_title("Deviation of deformation (volume)")
    ax3.set_ylabel("deform - true_deform")

    vref = np.linspace(volume.min(), volume.max(), 1000)
    result_deform_v = fit_exponential(dataf=delta_deform,
                                      datax=volume * pxscalev,
                                      x0=0,
                                      offset=0.0013,
                                      decaycoeffs=[40, 450, 6040])

    ax3.plot(volume, delta_deform, ".", ms=1, label="sphere")
    ax3.plot(volume, result_deform_v.best_fit, label="fit to sphere data")
    ax3.plot(vref, corr_deform_with_volume(vref), linestyle=(
        0, (5, 10)), color="k", alpha=.5, label="reference fit")

    # New figure volume-volume
    ax4.set_title("Deviation of volume (volume)")
    ax4.set_ylabel("true_volume / volume")

    result_volume = fit_exponential(dataf=delta_volume,
                                    datax=volume * pxscalev,
                                    x0=0,
                                    offset=1.023,
                                    decaycoeffs=[8.3, 102, 885, 8830])

    ax4.plot(volume, delta_volume, ".", ms=1, label="sphere")
    ax4.plot(volume, result_volume.best_fit, label="fit to sphere data")
    ax4.plot(vref, corr_volume_with_volume(vref), linestyle=(
        0, (5, 10)), color="k", alpha=.5, label="reference fit")


def plot_emodulus_deviation_dependency(path, ax5, ax6, ax7, ax8):
    """Visualize the effect of pixelation on emodulus

    Reproduces figure 5 of Herold's arXiv paper and goes
    beyond that.

    The idea is to add the pixelation correction (which would
    otherwise be subtracted) before interpolatimg the emodulus
    and then visualize by how far we are off compared to when
    we don't add the pixelation correction.

    Parameters
    ----------
    path: pathlib.Path
        Path to a FEM simulation results file. This is
        used for generating the deform-volume LUT
    ax5: matplotlib.Axis
        overestimation of emodulus for deform-vs-area_um
    ax6: matplotlib.Axis
        overestimation of emodulus for area_um-vs-area_um
    ax7: matplotlib.Axis
        overestimation of emodulus for deform-vs-volume
    ax8: matplotlib.Axis
        overestimation of emodulus for volume-vs-volume

    Notes
    -----
    - The relative error for area_um and volume are negative.
    - The low relative errors for area_um and volume tell us
      that we don't need a pixelation correction here. This
      also becomes obvious when looking at the LUT: A change
      in area (move horizontally through the LUT) does not
      change the emodulus much.
    - The default dclab LUT is used for deform-area_um data
    """
    # Plot expected deviations
    # deform-area_um
    ax5.set_ylabel("relative error (1-E/E0)")
    area_um = np.linspace(5, 290, 200)
    emodkw = dict(
        area_um=area_um,
        medium=15,
        channel_width=20,
        flow_rate=0.04,
        px_um=0,
        temperature=None)
    for dv in [0.005, 0.01, 0.02, 0.03, 0.1]:
        deform = np.ones_like(area_um) * dv
        emod0 = dclab.features.emodulus.get_emodulus(
            deform=deform, **emodkw)
        pxcorr = corr_deform_with_area_um(area_um)
        emod1 = dclab.features.emodulus.get_emodulus(
            deform=deform+pxcorr, **emodkw)
        ax5.plot(area_um, 1-emod1/emod0, label="{}".format(dv))

    # area_um-area_um
    ax6.set_ylabel("negative relative error -(1-E/E0)")
    emodkw = dict(
        medium=15,
        channel_width=20,
        flow_rate=0.04,
        px_um=0,
        temperature=None)
    for dv in [0.005, 0.01, 0.02, 0.03, 0.1]:
        deform = np.ones_like(area_um) * dv
        emod0 = dclab.features.emodulus.get_emodulus(
            deform=deform, area_um=area_um, **emodkw)
        pxcorr = corr_area_um_with_area_um(area_um)
        emod1 = dclab.features.emodulus.get_emodulus(
            deform=deform, area_um=area_um+pxcorr, **emodkw)
        ax6.plot(area_um, -(1-emod1/emod0), label="{}".format(dv))

    # deform-volume
    ax7.set_ylabel("relative error (1-E/E0)")
    lut_data_volume = get_lut_volume(path=path)
    volume = np.linspace(50, 3200, 200)
    emodkwv = dict(
        volume=volume,
        medium=15,
        channel_width=20,
        flow_rate=0.04,
        px_um=0,
        temperature=None,
        lut_data=lut_data_volume)
    for dv in [0.005, 0.01, 0.02, 0.03, 0.1]:
        deform = np.ones_like(volume) * dv
        emod0 = dclab.features.emodulus.get_emodulus(
            deform=deform, **emodkwv)
        pxcorr = corr_deform_with_volume(volume)
        emod1 = dclab.features.emodulus.get_emodulus(
            deform=deform+pxcorr, **emodkwv)
        ax7.plot(volume, 1-emod1/emod0, label="{}".format(dv))

    # volume-volume
    ax8.set_ylabel("negative relative error -(1-E/E0)")
    lut_data_volume = get_lut_volume(path=path)
    volume = np.linspace(50, 3200, 200)
    emodkwv = dict(
        medium=15,
        channel_width=20,
        flow_rate=0.04,
        px_um=0,
        temperature=None,
        lut_data=lut_data_volume)
    for dv in [0.005, 0.01, 0.02, 0.03, 0.1]:
        deform = np.ones_like(volume) * dv
        emod0 = dclab.features.emodulus.get_emodulus(
            deform=deform, volume=volume, **emodkwv)
        pxcorr = corr_volume_with_volume(volume)
        emod1 = dclab.features.emodulus.get_emodulus(
            deform=deform, volume=volume+pxcorr, **emodkwv)
        ax8.plot(volume, -(1-emod1/emod0), label="{}".format(dv))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("fempath",
                        help="Path to an .hdf5 FEM simulation file",
                        )
    args = parser.parse_args()
    path = pathlib.Path(args.fempath)

    fig = plt.figure(figsize=(15, 10))
    # plots for visualizing the pixelation effect
    ax1 = plt.subplot(241)
    ax2 = plt.subplot(242)
    ax3 = plt.subplot(243)
    ax4 = plt.subplot(244)
    # plots for visualizing the effect of pixelation on emodulus
    ax5 = plt.subplot(245)
    ax6 = plt.subplot(246)
    ax7 = plt.subplot(247)
    ax8 = plt.subplot(248)

    plot_numerical_data(path, ax1, ax2, ax3, ax4)
    plot_and_fit_disk_data(ax1, ax2, ax3, ax4)
    plot_emodulus_deviation_dependency(path, ax5, ax6, ax7, ax8)

    for ax in [ax1, ax2, ax5, ax6]:
        ax.set_xlim(0, 290)
        ax.set_xlabel("area_um [µm²]")

    for ax in [ax3, ax4, ax7, ax8]:
        ax.set_xlim(0, 3200)
        ax.set_xlabel("volume [µm³]")

    for ax in [ax1, ax2, ax3, ax4]:
        ax.legend()
        ax.grid()

    for ax in [ax5, ax6, ax7, ax8]:
        ax.grid()
        ax.legend(title="deformation")
        ax.set_ylim(1e-4, 1)
        ax.set_yscale("log")

    plt.tight_layout()
    plt.savefig(f"{path.stem}_pixelation_correction_gen_{PIXEL_SIZE}.png",
                dpi=150)
    plt.show()



from setuptools import Extension, setup

import numpy as np


setup(
    ext_modules=[
        Extension("dclab.downsampling",
                  sources=["dclab/downsampling.pyx"],
                  include_dirs=[np.get_include(), "_shared"]
                  ),
        Extension("dclab.external.skimage._shared.geometry",
                  sources=["dclab/external/skimage/_shared/geometry.pyx"],
                  include_dirs=[np.get_include()]
                  ),
        Extension("dclab.external.skimage._find_contours_cy",
                  sources=["dclab/external/skimage/_find_contours_cy.pyx"],
                  include_dirs=[np.get_include()]
                  ),
        Extension("dclab.external.skimage._pnpoly",
                  sources=["dclab/external/skimage/_pnpoly.pyx"],
                  include_dirs=[np.get_include(), "_shared"]
                  ),
        ]
)



"""Isoelastics management"""
import collections
from contextlib import ExitStack
import functools
# replace this import when dropping support for Python 3.8
# from importlib import resources as importlib_resources
import importlib_resources
import warnings

import numpy as np

from .. import definitions as dfn
from ..features import emodulus as feat_emod


class IsoelasticsEmodulusMeaninglessWarning(UserWarning):
    pass


class Isoelastics(object):
    def __init__(self, paths=None):
        """Isoelasticity line management

        Parameters
        ----------
        paths: list of pathlib.Path or list of str
            list of paths to files containing isoelasticity lines

        .. versionchanged:: 0.24.0
            The isoelasticity lines of the analytical model
            :cite:`Mietke2015` and the linear-elastic numerical
            model :cite:`Mokbel2017` were recomputed with an
            equidistant spacing. The metadata section of the text
            file format was restructured.
        """
        if paths is None:
            paths = []
        self._data = AutoRecursiveDict()

        for path in paths:
            self.load_data(path)

    def _add(self, isoel, col1, col2, lut_identifier, meta):
        """Convenience method for population self._data"""
        self._data[lut_identifier][col1][col2]["isoelastics"] = isoel
        self._data[lut_identifier][col1][col2]["meta"] = meta

        # Use advanced slicing to flip the data columns
        isoel_flip = [iso[:, [1, 0, 2]] for iso in isoel]
        self._data[lut_identifier][col2][col1]["isoelastics"] = isoel_flip
        self._data[lut_identifier][col2][col1]["meta"] = meta

    def add(self, isoel, col1, col2, channel_width,
            flow_rate, viscosity, method=None,
            lut_identifier=None):
        """Add isoelastics

        Parameters
        ----------
        isoel: list of ndarrays
            Each list item resembles one isoelastic line stored
            as an array of shape (N,3). The last column contains
            the emodulus data.
        col1: str
            Name of the first feature of all isoelastics
            (e.g. isoel[0][:,0])
        col2: str
            Name of the second feature of all isoelastics
            (e.g. isoel[0][:,1])
        channel_width: float
            Channel width in µm
        flow_rate: float
            Flow rate through the channel in µL/s
        viscosity: float
            Viscosity of the medium in mPa*s
        method: str
            The method used to compute the isoelastics
            DEPRECATED since 0.32.0. Please use
            `lut_identifier` instead.
        lut_identifier: str:
            Look-up table identifier used to identify which
            isoelasticity lines to show. The function
            :func:`get_available_identifiers` returns a list
            of available identifiers.

        Notes
        -----
        The following isoelastics are automatically added for
        user convenience:

        - isoelastics with `col1` and `col2` interchanged
        - isoelastics for circularity if deformation was given
        """
        lut_identifier = check_lut_identifier(lut_identifier, method)

        for col in [col1, col2]:
            if not dfn.scalar_feature_exists(col):
                raise ValueError("Not a valid feature name: {}".format(col))

        meta = [channel_width, flow_rate, viscosity]

        # Add the feature data
        self._add(isoel, col1, col2, lut_identifier, meta)

        # Also add the feature data for circularity
        if "deform" in [col1, col2]:
            col1c, col2c = col1, col2
            if col1c == "deform":
                deform_ax = 0
                col1c = "circ"
            else:
                deform_ax = 1
                col2c = "circ"
            iso_circ = []
            for iso in isoel:
                iso = iso.copy()
                iso[:, deform_ax] = 1 - iso[:, deform_ax]
                iso_circ.append(iso)
            self._add(iso_circ, col1c, col2c, lut_identifier, meta)

    @staticmethod
    def add_px_err(isoel, col1, col2, px_um, inplace=False):
        """Undo pixelation correction

        Since isoelasticity lines are usually computed directly from
        the simulation data (e.g. the contour data are not discretized
        on a grid but are extracted from FEM simulations), they are
        not affected by pixelation effects as described in
        :cite:`Herold2017`.

        If the isoelasticity lines are displayed alongside experimental
        data (which are affected by pixelation effects), then the lines
        must be "un"-corrected, i.e. the pixelation error must be added
        to the lines to match the experimental data.

        Parameters
        ----------
        isoel: list of 2d ndarrays of shape (N, 3)
            Each item in the list corresponds to one isoelasticity
            line. The first column is defined by `col1`, the second
            by `col2`, and the third column is the emodulus.
        col1, col2: str
            Define the fist two columns of each isoelasticity line.
        px_um: float
            Pixel size [µm]
        inplace: bool
            If True, do not create a copy of the data in `isoel`
        """
        new_isoel = []
        for iso in isoel:
            iso = np.array(iso, copy=not inplace)
            delt1, delt2 = feat_emod.get_pixelation_delta_pair(
                feat1=col1, feat2=col2, data1=iso[:, 0], data2=iso[:, 1],
                px_um=px_um)
            iso[:, 0] += delt1
            iso[:, 1] += delt2
            new_isoel.append(iso)
        return new_isoel

    @staticmethod
    def convert(isoel, col1, col2,
                channel_width_in, channel_width_out,
                flow_rate_in, flow_rate_out,
                viscosity_in, viscosity_out,
                inplace=False):
        """Perform isoelastics scale conversion

        Parameters
        ----------
        isoel: list of 2d ndarrays of shape (N, 3)
            Each item in the list corresponds to one isoelasticity
            line. The first column is defined by `col1`, the second
            by `col2`, and the third column is the emodulus.
        col1, col2: str
            Define the fist to columns of each isoelasticity line.
            One of ["area_um", "circ", "deform"]
        channel_width_in: float
            Original channel width [µm]
        channel_width_out: float
            Target channel width [µm]
        flow_rate_in: float
            Original flow rate [µL/s]
        flow_rate_out: float
            Target flow rate [µL/s]
        viscosity_in: float
            Original viscosity [mPa*s]
        viscosity_out: float
            Target viscosity [mPa*s]
        inplace: bool
            If True, do not create a copy of the data in `isoel`

        Returns
        -------
        isoel_scale: list of 2d ndarrays of shape (N, 3)
            The scale-converted isoelasticity lines.

        Notes
        -----
        If only the positions of the isoelastics are of interest and
        not the value of the elastic modulus, then it is sufficient
        to supply values for the channel width and set the values
        for flow rate and viscosity to a constant (e.g. 1).

        See Also
        --------
        dclab.features.emodulus.scale_linear.scale_feature: scale
            conversion method used
        """
        new_isoel = []

        for iso in isoel:
            iso = np.array(iso, copy=not inplace)
            scale_kw = {"channel_width_in": channel_width_in,
                        "channel_width_out": channel_width_out,
                        "flow_rate_in": flow_rate_in,
                        "flow_rate_out": flow_rate_out,
                        "viscosity_in": viscosity_in,
                        "viscosity_out": viscosity_out,
                        "inplace": True}
            feat_emod.scale_feature(col1, data=iso[:, 0], **scale_kw)
            feat_emod.scale_feature(col2, data=iso[:, 1], **scale_kw)
            feat_emod.scale_emodulus(emodulus=iso[:, 2], **scale_kw)
            new_isoel.append(iso)
        return new_isoel

    def get(self, col1, col2, channel_width, method=None, lut_identifier=None,
            flow_rate=None, viscosity=None, add_px_err=False, px_um=None):
        """Get isoelastics

        Parameters
        ----------
        col1: str
            Name of the first feature of all isoelastics
            (e.g. isoel[0][:,0])
        col2: str
            Name of the second feature of all isoelastics
            (e.g. isoel[0][:,1])
        channel_width: float
            Channel width in µm
        method: str
            The method used to compute the isoelastics
            DEPRECATED since 0.32.0. Please use
            `lut_identifier` instead.
        lut_identifier: str:
            Look-up table identifier used to identify which
            isoelasticity lines to show. The function
            :func:`get_available_identifiers` returns a list
            of available identifiers.
        flow_rate: float or `None`
            Flow rate through the channel in µL/s. If set to
            `None`, the flow rate of the imported data will
            be used (only do this if you do not need the
            correct values for elastic moduli).
        viscosity: float or `None`
            Viscosity of the medium in mPa*s. If set to
            `None`, the flow rate of the imported data will
            be used (only do this if you do not need the
            correct values for elastic moduli).
        add_px_err: bool
            If True, add pixelation errors according to
            C. Herold (2017), https://arxiv.org/abs/1704.00572
            and scripts/pixelation_correction.py
        px_um: float
            Pixel size [µm], used for pixelation error computation

        See Also
        --------
        dclab.features.emodulus.scale_linear.scale_feature: scale
            conversion method used
        dclab.features.emodulus.pxcorr.get_pixelation_delta:
            pixelation correction (applied to the feature data)
        """
        lut_identifier = check_lut_identifier(lut_identifier, method)

        for col in [col1, col2]:
            if not dfn.scalar_feature_exists(col):
                raise ValueError("Not a valid feature name: {}".format(col))

        if "isoelastics" not in self._data[lut_identifier][col2][col1]:
            msg = "No isoelastics matching {}, {}, {}".format(col1, col2,
                                                              lut_identifier)
            raise KeyError(msg)

        isoel = self._data[lut_identifier][col1][col2]["isoelastics"]
        meta = self._data[lut_identifier][col1][col2]["meta"]

        if flow_rate is None:
            flow_rate = meta[1]

        if viscosity is None:
            viscosity = meta[2]

        isoel_ret = self.convert(isoel, col1, col2,
                                 channel_width_in=meta[0],
                                 channel_width_out=channel_width,
                                 flow_rate_in=meta[1],
                                 flow_rate_out=flow_rate,
                                 viscosity_in=meta[2],
                                 viscosity_out=viscosity,
                                 inplace=False)

        if add_px_err:
            self.add_px_err(isoel=isoel_ret,
                            col1=col1,
                            col2=col2,
                            px_um=px_um,
                            inplace=True)

        return isoel_ret

    def get_with_rtdcbase(self, col1, col2, dataset, method=None,
                          lut_identifier=None, viscosity=None,
                          add_px_err=False):
        """Convenience method that extracts the metadata from RTDCBase

        Parameters
        ----------
        col1: str
            Name of the first feature of all isoelastics
            (e.g. isoel[0][:,0])
        col2: str
            Name of the second feature of all isoelastics
            (e.g. isoel[0][:,1])
        method: str
            The method used to compute the isoelastics
            DEPRECATED since 0.32.0. Please use
            `lut_identifier` instead.
        lut_identifier: str:
            Look-up table identifier used to identify which
            isoelasticity lines to show. The function
            :func:`get_available_identifiers` returns a list
            of available identifiers.
        dataset: dclab.rtdc_dataset.RTDCBase
            The dataset from which to obtain the metadata.
        viscosity: float, `None`, or False
            Viscosity of the medium in mPa*s. If set to
            `None`, the viscosity is computed from the meta
            data (medium, flow rate, channel width, temperature)
            in the [setup] config section. If this is not possible,
            the flow rate of the imported data is used and a warning
            will be issued.
        add_px_err: bool
            If True, add pixelation errors according to
            C. Herold (2017), https://arxiv.org/abs/1704.00572
            and scripts/pixelation_correction.py
        """
        lut_identifier = check_lut_identifier(lut_identifier, method)

        cfg = dataset.config
        if viscosity is None:
            if "temperature" in cfg["setup"] and "medium" in cfg["setup"]:
                viscosity = feat_emod.get_viscosity(
                    medium=cfg["setup"]["medium"],
                    channel_width=cfg["setup"]["channel width"],
                    flow_rate=cfg["setup"]["flow rate"],
                    temperature=cfg["setup"]["temperature"],
                    model='buyukurganci-2022',
                )
            else:
                warnings.warn("Computing emodulus data for isoelastics from "
                              + "RTDCBase is not possible. Isoelastics will "
                              + "not have correct emodulus values (this is "
                              + "not relevant for plotting).",
                              IsoelasticsEmodulusMeaninglessWarning)
        return self.get(col1=col1,
                        col2=col2,
                        lut_identifier=lut_identifier,
                        channel_width=cfg["setup"]["channel width"],
                        flow_rate=cfg["setup"]["flow rate"],
                        viscosity=viscosity,
                        add_px_err=add_px_err,
                        px_um=cfg["imaging"]["pixel size"])

    def load_data(self, path):
        """Load isoelastics from a text file

        Parameters
        ----------
        path: str or pathlib.Path
            Path to an isoelasticity lines text file
        """
        isodata, meta = feat_emod.load_mtext(path)
        assert len(meta["column features"]) == 3
        assert meta["column features"][2] == "emodulus"
        assert meta["lut identifier"]

        # Slice out individual isoelastics
        emoduli = np.unique(isodata[:, 2])
        isoel = []
        for emod in emoduli:
            where = isodata[:, 2] == emod
            isoel.append(isodata[where])

        # Add isoelastics to instance
        self.add(isoel=isoel,
                 col1=meta["column features"][0],
                 col2=meta["column features"][1],
                 channel_width=meta["channel_width"],
                 flow_rate=meta["flow_rate"],
                 viscosity=meta["fluid_viscosity"],
                 lut_identifier=meta["lut identifier"])


class AutoRecursiveDict(collections.UserDict):
    def __getitem__(self, key):
        if key not in self:
            self[key] = AutoRecursiveDict()
        return super(AutoRecursiveDict, self).__getitem__(key)


def check_lut_identifier(lut_identifier, method):
    """Transitional function that can be removed once `method` is removed"""
    if lut_identifier is None:
        if method is not None:
            warnings.warn("The `method` argument is deprecated "
                          + "please use `lut_identifier` instead!",
                          DeprecationWarning)
            if method == "analytical":
                lut_identifier = "LE-2D-ana-18"
            elif method == "numerical":
                lut_identifier = "LE-2D-FEM-19"
            else:
                raise ValueError("Please read the docstring")
    # Now check again (this can be removed once lut_identifier becomes
    # a non-keyword argument)
    if lut_identifier is None:
        raise ValueError("Please specify `lut_identifier`!")
    return lut_identifier


@functools.lru_cache()
def get_available_identifiers():
    """Return a list of available LUT identifiers"""
    isofiles = get_available_files()
    ids = []
    for pp in isofiles:
        _, meta = feat_emod.load_mtext(pp)
        ids.append(meta["lut identifier"])
    return sorted(set(ids))


@functools.lru_cache()
def get_available_files():
    """Return list of available isoelasticity line files in dclab"""
    isofiles = []
    for pp in importlib_resources.files('dclab.isoelastics').iterdir():
        name = pp.name
        if name.startswith("iso") and name.endswith(".txt"):
            isofiles.append(name)
    return sorted(isofiles)


def get_default():
    """Return default isoelasticity lines"""
    # name of the isoelastics files
    isofiles = get_available_files()
    # file manager to temporarily work with the package files
    file_manager = ExitStack()
    paths = []
    for name in isofiles:
        ref = importlib_resources.files('dclab.isoelastics') / name
        path = file_manager.enter_context(importlib_resources.as_file(ref))
        paths.append(path)
    # this loads the data into memory, so we don't need the files afterward
    isoels = Isoelastics(paths=paths)
    # close the file manager, deleting all files
    file_manager.close()
    return isoels



import dclab
from dclab import cli, new_dataset, rtdc_dataset, RTDCWriter

import h5py
import numpy as np
import pytest

from helper_methods import retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_split():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    paths = cli.split(path_in=path_in, split_events=3, ret_out_paths=True)
    with new_dataset(path_in) as ds:
        ecount = 0
        for pp in paths:
            with new_dataset(pp) as di:
                for feat in ds.features_scalar:
                    if feat in ["index",
                                "time",  # issue 204
                                ]:
                        continue
                    assert np.all(
                        ds[feat][ecount:ecount + len(di)] == di[feat]), feat
                ecount += len(di)


def test_split_mapped_basin():
    """Splitting should produce mapped basins"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "image" in ds.features

    # same directory (will be cleaned up with path_in)
    path_out = h5path.parent

    # split it
    paths = cli.split(path_in=h5path_small,
                      path_out=path_out,
                      split_events=3,
                      ret_out_paths=True)

    with dclab.new_dataset(h5path) as ds0:
        # sanity check
        assert not np.all(ds0["image"][:3] == ds0["image"][3:6])
        for ii, pp in enumerate(paths):
            with dclab.new_dataset(pp) as dsi:
                # sanity check for existence of image
                assert "image" not in dsi.features_innate
                assert "image" in dsi.features_basin
                # Make sure the mapping in the split files is correct
                assert np.all(dsi["image"][:] == ds0["image"][ii*3:(ii+1)*3])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_split_compressed():
    """Make sure the split output data are compressed"""
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.parent

    paths = cli.split(path_in=path_in,
                      path_out=path_out,
                      split_events=3,
                      ret_out_paths=True)

    for pp in paths:
        ic = rtdc_dataset.check.IntegrityChecker(pp)
        ccue = ic.check_compression()[0]
        assert ccue.data["uncompressed"] == 0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_split_traces():
    path_in = retrieve_data("fmt-hdf5_fl_2018.zip")
    paths = cli.split(path_in=path_in, split_events=3, ret_out_paths=True)
    with new_dataset(path_in) as ds:
        ecount = 0
        for pp in paths:
            with new_dataset(pp) as di:
                for flkey in ds["trace"].keys():
                    trace1 = ds["trace"][flkey][ecount:ecount + len(di)]
                    trace2 = di["trace"][flkey][:]
                    assert len(trace1) == len(trace2)
                    assert np.all(trace1 == trace2), flkey
                ecount += len(di)
